@@ -15,14 +15,210 @@
 #include <apt-pkg/algorithms.h>
 #include <apt-pkg/error.h>
 #include <apt-pkg/pkgcache.h>
+#include <apt-pkg/pkgsystem.h>
+#include <apt-pkg/policy.h>
+#include <apt-pkg/version.h>
+
+#include <stdlib.h>
+
+#include <cwidget/generic/util/transcode.h>
+
+namespace cw = cwidget;
+
+namespace
+{
+  bool cmdline_do_build_depends(const string &pkg,
+				std::set<pkgCache::PkgIterator> &seen_virtual_packages,
+				cmdline_version_source version_source,
+				const string &version_source_string,
+				bool arch_only,
+				pkgPolicy &policy,
+				pkgset &to_install, pkgset &to_hold,
+				pkgset &to_remove, pkgset &to_purge,
+				int verbose,
+				bool allow_auto)
+  {
+    aptitude::cmdline::source_package sourcepkg =
+      aptitude::cmdline::find_source_package(pkg,
+					     version_source,
+					     version_source_string);
+
+    if(!sourcepkg.valid())
+      {
+	printf(_("Unable to find the source package for \"%s\".\n"),
+	       pkg.c_str());
+	return false;
+      }
+
+    if(apt_cache_file == NULL)
+      {
+	// Should never happen.
+	printf("Sanity-check failed: apt_cache_file should not be NULL here.\n");
+	return false;
+      }
+
+    bool rval = true;
+    bool last_was_or = false;
+    typedef pkgSrcRecords::Parser::BuildDepRec BuildDepRec;
+    typedef std::vector<BuildDepRec> BuildDepList;
+    for(BuildDepList::const_iterator it = sourcepkg.get_build_deps().begin();
+	it != sourcepkg.get_build_deps().end(); ++it)
+      {
+	bool is_conflict = (it->Type == pkgSrcRecords::Parser::BuildConflictIndep ||
+			    it->Type == pkgSrcRecords::Parser::BuildConflict);
+
+	if(!arch_only && (it->Type == pkgSrcRecords::Parser::BuildDependIndep ||
+			  it->Type == pkgSrcRecords::Parser::BuildConflictIndep))
+	  continue;
+
+	BuildDepList::const_iterator or_group_start = it;
+
+	// Find the bounds of the OR group.
+	while((it->Op & pkgCache::Dep::Or) != 0)
+	  ++it;
+
+	bool ok = true;
+	if(!is_conflict)
+	  {
+	    // Walk over the OR group and see if it's satisfied.
+	    //
+	    // NB: if two build-deps require incompatible versions
+	    // of the same package, we'll just get confused.
+	    bool satisfied = false;
+	    for(BuildDepList::const_iterator it2 = or_group_start;
+		!satisfied && it2 <= it; ++it2)
+	      {
+		pkgCache::PkgIterator pkg = (*apt_cache_file)->FindPkg(it2->Package);
+		if(!pkg.end())
+		  {
+		    pkgCache::VerIterator ver = pkg.CurrentVer();
+		    if(!ver.end() &&
+		       _system->VS->CheckDep(ver.VerStr(),
+					     it2->Op,
+					     it2->Version.c_str()))
+		       satisfied = true;
+		  }
+	      }
+
+	    if(!satisfied)
+	      {
+		for(BuildDepList::const_iterator it2 = or_group_start;
+		    !satisfied && it2 <= it; ++it2)
+		  {
+		    pkgCache::PkgIterator pkg = (*apt_cache_file)->FindPkg(it2->Package);
+		    if(!pkg.end())
+		      {
+			int best_priority;
+			pkgCache::VerIterator best_ver;
+			for(pkgCache::VerIterator ver = pkg.VersionList();
+			    !ver.end(); ++ver)
+			  {
+			    if(!_system->VS->CheckDep(ver.VerStr(), it->Op, it->Version.c_str()))
+			      continue;
+
+			    for(pkgCache::VerFileIterator vf = ver.FileList();
+				!vf.end(); ++vf)
+			      {
+				const int priority = policy.GetPriority(vf.File());
+				if(best_ver.end() || best_priority < priority)
+				  {
+				    best_ver = ver;
+				    best_priority = priority;
+				  }
+			      }
+			  }
+
+			if(!best_ver.end())
+			  {
+			    cmdline_applyaction(cmdline_install,
+						pkg,
+						seen_virtual_packages,
+						to_install, to_hold,
+						to_remove, to_purge,
+						verbose,
+						cmdline_version_version,
+						best_ver.VerStr(),
+						policy,
+						arch_only,
+						allow_auto);
+			    satisfied = true;
+			  }
+		      }
+		  }
+
+		if(!satisfied)
+		  ok = false;
+	      }
+	  }
+	else // if(!is_conflict); now we're satisfying a conflict.
+	  {
+	    for(BuildDepList::const_iterator it2 = or_group_start;
+		it2 <= it; ++it2)
+	      {
+		pkgCache::PkgIterator pkg = (*apt_cache_file)->FindPkg(it2->Package);
+		if(!pkg.end())
+		  {
+		    pkgCache::VerIterator ver = pkg.CurrentVer();
+		    if(!ver.end() &&
+		       _system->VS->CheckDep(ver.VerStr(),
+					     it2->Op,
+					     it2->Version.c_str()))
+		      {
+			cmdline_applyaction(cmdline_remove,
+					    pkg,
+					    seen_virtual_packages,
+					    to_install, to_hold,
+					    to_remove, to_purge,
+					    verbose,
+					    cmdline_version_cand,
+					    std::string(),
+					    policy,
+					    arch_only,
+					    allow_auto);
+		      }
+		  }
+	      }
+	  }
+
+	if(!ok)
+	  {
+	    rval = false;
+
+	    std::string build_dep_description = pkgSrcRecords::Parser::BuildDepType(it->Type);;
+	    for(BuildDepList::const_iterator it2 = or_group_start;
+		it2 <= it; ++it2)
+	      {
+		if(it2 != or_group_start)
+		  build_dep_description += " | ";
+		build_dep_description += it2->Package;
+		if((it2->Op & ~pkgCache::Dep::Or) != pkgCache::Dep::NoOp)
+		  {
+		    build_dep_description += " (";
+		    build_dep_description += pkgCache::CompType(it2->Type);
+		    build_dep_description += " ";
+		    build_dep_description += it2->Version;
+		    build_dep_description += ")";
+		  }
+	      }
+	    printf(_("Unable to satisfy the build-depends: %s."),
+		   build_dep_description.c_str());
+	  }
+      }
+
+    return rval;
+  }
+}
 
 bool cmdline_applyaction(cmdline_pkgaction_type action,
 			 pkgCache::PkgIterator pkg,
+			 std::set<pkgCache::PkgIterator> &seen_virtual_packages,
 			 pkgset &to_install, pkgset &to_hold,
 			 pkgset &to_remove, pkgset &to_purge,
 			 int verbose,
 			 cmdline_version_source source,
 			 const string &sourcestr,
+			 pkgPolicy &policy,
+			 bool arch_only,
 			 bool allow_auto)
 {
   // Handle virtual packages.
@@ -30,24 +226,32 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
     {
       if(pkg.VersionList().end())
 	{
-	  for(pkgCache::PrvIterator prv=pkg.ProvidesList(); !prv.end(); ++prv)
+	  const bool seen_in_first_pass =
+	    seen_virtual_packages.find(pkg) != seen_virtual_packages.end();
+
+	  if(!seen_in_first_pass)
 	    {
-	      if(prv.OwnerPkg().CurrentVer()==prv.OwnerVer())
+	      for(pkgCache::PrvIterator prv=pkg.ProvidesList(); !prv.end(); ++prv)
 		{
-		  if(verbose>0)
-		    printf(_("Note: \"%s\", providing the virtual package\n      \"%s\", is already installed.\n"),
-			   prv.OwnerPkg().Name(), pkg.Name());
-		  return true;
-		}
-	      else if((*apt_cache_file)[prv.OwnerPkg()].InstVerIter(*apt_cache_file)==prv.OwnerVer())
-		{
-		  if(verbose>0)
-		    printf(_("Note: \"%s\", providing the virtual package\n      \"%s\", is already going to be installed.\n"),
-			   prv.OwnerPkg().Name(), pkg.Name());
-		  return true;
+		  if(prv.OwnerPkg().CurrentVer()==prv.OwnerVer())
+		    {
+		      if(verbose>0)
+			printf(_("Note: \"%s\", providing the virtual package\n      \"%s\", is already installed.\n"),
+			       prv.OwnerPkg().Name(), pkg.Name());
+		      return true;
+		    }
+		  else if((*apt_cache_file)[prv.OwnerPkg()].InstVerIter(*apt_cache_file)==prv.OwnerVer())
+		    {
+		      if(verbose>0)
+			printf(_("Note: \"%s\", providing the virtual package\n      \"%s\", is already going to be installed.\n"),
+			       prv.OwnerPkg().Name(), pkg.Name());
+		      return true;
+		    }
 		}
 	    }
 
+	  seen_virtual_packages.insert(pkg);
+
 	  // See if there's only one possible package to install.
 	  pkgvector cands;
 
@@ -60,24 +264,31 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 
 	  if(cands.size()==0)
 	    {
-	      printf(_("\"%s\" exists in the package database, but it is not a\nreal package and no package provides it.\n"),
+	      if(!seen_in_first_pass)
+		printf(_("\"%s\" exists in the package database, but it is not a\nreal package and no package provides it.\n"),
 		     pkg.Name());
 	      return false;
 	    }
 	  else if(cands.size()>1)
 	    {
-	      printf(_("\"%s\" is a virtual package provided by:\n"),
-		     pkg.Name());
-	      cmdline_show_pkglist(cands);
-	      printf(_("You must choose one to install.\n"));
+	      if(!seen_in_first_pass)
+		{
+		  printf(_("\"%s\" is a virtual package provided by:\n"),
+			 pkg.Name());
+		  cmdline_show_pkglist(cands);
+		  printf(_("You must choose one to install.\n"));
+		}
 	      return false;
 	    }
 	  else if(cands.size()==1)
-	      {
-		printf(_("Note: selecting \"%s\" instead of the\n      virtual package \"%s\"\n"),
-		       cands[0].Name(), pkg.Name());
-		pkg=cands[0];
-	      }
+	    {
+	      if(!seen_in_first_pass)
+		{
+		  printf(_("Note: selecting \"%s\" instead of the\n      virtual package \"%s\"\n"),
+			 cands[0].Name(), pkg.Name());
+		}
+	      pkg = cands[0];
+	    }
  	}
     }
 
@@ -164,7 +375,8 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
       (*apt_cache_file)->mark_keep(pkg, false, false, NULL);
       break;
     case cmdline_unhold:
-      (*apt_cache_file)->get_ext_state(pkg).selection_state = pkgCache::State::Install;
+      if((*apt_cache_file)[pkg].Keep())
+	(*apt_cache_file)->mark_keep(pkg, false, false, NULL);
       break;
     case cmdline_markauto:
       (*apt_cache_file)->mark_auto_installed(pkg, true, NULL);
@@ -183,6 +395,17 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 	    (*apt_cache_file)->forbid_upgrade(pkg, candver.VerStr(), NULL);
 	}
       break;
+    case cmdline_build_depends:
+      return cmdline_do_build_depends(pkg.Name(),
+				      seen_virtual_packages,
+				      source,
+				      sourcestr,
+				      arch_only,
+				      policy,
+				      to_install, to_hold,
+				      to_remove, to_purge,
+				      verbose,
+				      allow_auto);
     default:
       fprintf(stderr, "Internal error: impossible pkgaction type\n");
       abort();
@@ -192,12 +415,16 @@ bool cmdline_applyaction(cmdline_pkgaction_type action,
 }
 
 bool cmdline_applyaction(string s,
+			 std::set<pkgCache::PkgIterator> &seen_virtual_packages,
 			 cmdline_pkgaction_type action,
 			 pkgset &to_install, pkgset &to_hold,
 			 pkgset &to_remove, pkgset &to_purge,
 			 int verbose,
+			 pkgPolicy &policy, bool arch_only,
 			 bool allow_auto)
 {
+  using namespace aptitude::matching;
+
   bool rval=true;
 
   cmdline_version_source source=cmdline_version_cand;
@@ -215,15 +442,18 @@ bool cmdline_applyaction(string s,
       for(pkgCache::PkgIterator pkg=(*apt_cache_file)->PkgBegin();
 	  !pkg.end(); ++pkg)
 	{
-	  std::list<std::string> *tasks=get_tasks(pkg);
+	  std::set<std::string> *tasks = get_tasks(pkg);
 
-	  for(std::list<std::string>::iterator i=tasks->begin();
+	  for(std::set<std::string>::iterator i = tasks->begin();
 	      i!=tasks->end(); ++i)
 	    if(*i==s)
 	      rval=cmdline_applyaction(action, pkg,
+				       seen_virtual_packages,
 				       to_install, to_hold, to_remove, to_purge,
 				       verbose, source,
-				       sourcestr, allow_auto) && rval;
+				       sourcestr,
+				       policy, arch_only,
+				       allow_auto) && rval;
 	}
 
       // break out.
@@ -237,7 +467,8 @@ bool cmdline_applyaction(string s,
   if(source == cmdline_version_version &&
      action != cmdline_install &&
      action != cmdline_forbid_version &&
-     action != cmdline_installauto)
+     action != cmdline_installauto &&
+     action != cmdline_build_depends)
     {
       printf(_("You can only specify a package version with an 'install' command or a 'forbid-version' command.\n"));
       return false;
@@ -245,38 +476,42 @@ bool cmdline_applyaction(string s,
 
   if(source == cmdline_version_archive &&
      action != cmdline_install &&
-     action != cmdline_installauto)
+     action != cmdline_installauto &&
+     action != cmdline_build_depends)
     {
       printf(_("You can only specify a package archive with an 'install' command.\n"));
       return false;
     }
 
-  if(package.find('~')==package.npos)
+  if(!cmdline_is_search_pattern(package))
     {
       pkgCache::PkgIterator pkg=(*apt_cache_file)->FindPkg(package.c_str());
       if(pkg.end())
 	{
+	  // Assume the user asked for a source package.
+	  if(action == cmdline_build_depends)
+	    return cmdline_do_build_depends(package,
+					    seen_virtual_packages,
+					    source,
+					    sourcestr,
+					    arch_only,
+					    policy,
+					    to_install, to_hold,
+					    to_remove, to_purge,
+					    verbose,
+					    allow_auto);
+
 	  // Maybe they misspelled the package name?
 	  pkgvector possible;
-	  pkg_matcher *m=parse_pattern(package);
-
-	  if(!m)
-	    {
-	      _error->Error("Badly formed pattern %s", package.c_str());
-	      _error->DumpErrors();
-
-	      return false;
-	    }
 
 	  for(pkgCache::PkgIterator j=(*apt_cache_file)->PkgBegin();
 	      !j.end(); ++j)
 	    {
-	      if(!(j.VersionList().end() && j.ProvidesList().end()) && m->matches(j, *apt_cache_file, *apt_package_records))
+	      if(!(j.VersionList().end() && j.ProvidesList().end()) &&
+		 strstr(j.Name(), package.c_str()) != NULL)
 		possible.push_back(j);
 	    }
 
-	  delete m;
-
 	  if(!possible.empty())
 	    {
 	      // Don't overwhelm the user.
@@ -290,23 +525,21 @@ bool cmdline_applyaction(string s,
 	    }
 	  else
 	    {
-	      m=parse_pattern("~d"+s);
-
-	      if(!m)
-		{
-		  _error->DumpErrors();
-		  return false;
-		}
-
 	      for(pkgCache::PkgIterator j=(*apt_cache_file)->PkgBegin();
 		  !j.end(); ++j)
 		{
-		  if(m->matches(j, *apt_cache_file, *apt_package_records))
-		    possible.push_back(j);
+		  for(pkgCache::VerIterator v = j.VersionList();
+		      !v.end(); ++v)
+		    {
+		      std::wstring desc = get_long_description(v, apt_package_records);
+		      if(desc.find(cw::util::transcode(package)) != desc.npos)
+			{
+			  possible.push_back(j);
+			  break;
+			}
+		    }
 		}
 
-	      delete m;
-
 	      if(possible.empty())
 		printf(_("Couldn't find any package whose name or description matched \"%s\"\n"), package.c_str());
 	      else if(possible.size()>40)
@@ -321,14 +554,15 @@ bool cmdline_applyaction(string s,
 	  return false;
 	}
 
-      rval=cmdline_applyaction(action, pkg,
-			       to_install, to_hold, to_remove, to_purge,
-			       verbose, source,
-			       sourcestr, allow_auto);
+      rval = cmdline_applyaction(action, pkg,
+				 seen_virtual_packages,
+				 to_install, to_hold, to_remove, to_purge,
+				 verbose, source,
+				 sourcestr, policy, arch_only, allow_auto);
     }
   else
     {
-      pkg_matcher *m=parse_pattern(package.c_str());
+      pkg_matcher *m = parse_pattern(package.c_str());
       if(!m)
 	{
 	  _error->DumpErrors();
@@ -340,11 +574,14 @@ bool cmdline_applyaction(string s,
 	{
 	  pkgCache::VerIterator testver;
 
-	  if(m->matches(pkg, *apt_cache_file, *apt_package_records))
-	    rval=cmdline_applyaction(action, pkg,
-				     to_install, to_hold, to_remove, to_purge,
-				     verbose, source,
-				     sourcestr, allow_auto) && rval;
+	  if(apply_matcher(m, pkg, *apt_cache_file, *apt_package_records))
+	    rval = cmdline_applyaction(action, pkg,
+				       seen_virtual_packages,
+				       to_install, to_hold, to_remove, to_purge,
+				       verbose, source,
+				       sourcestr,
+				       policy, arch_only,
+				       allow_auto) && rval;
 	}
 
       delete m;
@@ -394,6 +631,18 @@ static bool parse_action_str(const string &s,
 	  if(loc<s.size())
 	    switch(s[loc])
 	      {
+	      case 'B':
+		// &BD to install build-deps.
+		++loc;
+		if(loc < s.size() && s[loc] == 'D')
+		  {
+		    action = cmdline_build_depends;
+		    ++loc;
+		  }
+		else
+		  return false;
+
+		break;
 	      case 'M':
 		action=cmdline_markauto;
 		++loc;
@@ -416,9 +665,11 @@ static bool parse_action_str(const string &s,
 }
 
 void cmdline_parse_action(string s,
+			  std::set<pkgCache::PkgIterator> &seen_virtual_packages,
 			  pkgset &to_install, pkgset &to_hold,
 			  pkgset &to_remove, pkgset &to_purge,
 			  int verbose,
+			  pkgPolicy &policy, bool arch_only,
 			  bool allow_auto)
 {
   string::size_type loc=0;
@@ -450,10 +701,13 @@ void cmdline_parse_action(string s,
 	      while(loc<s.size() && !isspace(s[loc]))
 		pkgname+=s[loc++];
 
-	      if(!cmdline_applyaction(pkgname, action,
+	      if(!cmdline_applyaction(pkgname,
+				      seen_virtual_packages,
+				      action,
 				      to_install, to_hold,
 				      to_remove, to_purge,
-				      verbose, allow_auto))
+				      verbose, policy,
+				      arch_only, allow_auto))
 		return;
 	    }
 	}
