@@ -8,6 +8,7 @@
 #include "cmdline_changelog.h"
 #include "cmdline_resolver.h"
 #include "cmdline_show.h"
+#include "cmdline_show_broken.h"
 #include "cmdline_util.h"
 #include "cmdline_why.h"
 
@@ -17,6 +18,7 @@
 #include <generic/apt/config_signal.h>
 #include <generic/apt/download_signal_log.h>
 #include <generic/apt/infer_reason.h>
+#include <generic/apt/matchers.h>
 
 #include <generic/util/util.h>
 
@@ -30,6 +32,8 @@
 #include <apt-pkg/sourcelist.h>
 #include <apt-pkg/strutl.h>
 
+#include <algorithm>
+
 using namespace std;
 namespace cw = cwidget;
 
@@ -99,6 +103,211 @@ static string reason_string_list(set<reason> &reasons)
   return s;
 }
 
+namespace
+{
+  // Sort action vectors by the name of the package in the first
+  // action.
+  struct compare_first_action
+  {
+    typedef aptitude::why::action action;
+    bool operator()(const std::vector<action> &reason1,
+		    const std::vector<action> &reason2)
+    {
+      if(reason1.empty())
+	return !reason2.empty();
+      else if(reason2.empty())
+	return false;
+      else
+	return strcmp(reason1.front().get_dep().ParentPkg().Name(),
+		      reason2.front().get_dep().ParentPkg().Name());
+    }
+  };
+
+  std::string roots_string(const pkgCache::PkgIterator &pkg,
+			   int verbose)
+  {
+    using namespace aptitude::why;
+    pkgDepCache::StateCache &state((*apt_cache_file)[pkg]);
+
+    // Don't show anything for packages that are kept back or are
+    // manually installed.
+    if(state.Keep() ||
+       (state.Install() && ((state.Flags & pkgCache::Flag::Auto) == 0)))
+      return "";
+
+    target t(state.Install() ? target::Install(pkg) : target::Remove(pkg));
+    std::vector<aptitude::matching::pkg_matcher *> leaves;
+    leaves.push_back(aptitude::matching::parse_pattern("?not(?automatic)"));
+
+    std::vector<std::vector<action> > reasons;
+
+    std::vector<search_params> params;
+    params.push_back(search_params(search_params::InstallNotCurrent,
+				   search_params::DependsOnly,
+				   false));
+    params.push_back(search_params(search_params::InstallNotCurrent,
+				   search_params::DependsOnly,
+				   true));
+    params.push_back(search_params(search_params::InstallNotCurrent,
+				   search_params::Recommends,
+				   false));
+    params.push_back(search_params(search_params::InstallNotCurrent,
+				   search_params::Recommends,
+				   true));
+
+    for(std::vector<search_params>::const_iterator it = params.begin();
+	it != params.end(); ++it)
+      {
+	if(find_justification(t,
+			      leaves,
+			      *it,
+			      true,
+			      reasons))
+	  break;
+      }
+
+    if(reasons.size() == 0)
+      return "";
+
+    if(verbose == 0)
+      {
+	std::set<std::string> root_names;
+	for(std::vector<std::vector<action> >::const_iterator it =
+	      reasons.begin(); it != reasons.end(); ++it)
+	  {
+	    // Shouldn't happen, but deal anyway.
+	    if(it->empty())
+	      continue; // Generate an internal error here?
+
+	    const action &act(it->front());
+
+	    // This can happen in the case of an impure virtual
+	    // package.  e.g., A is manually installed and B provides
+	    // A.
+	    if(act.get_dep().end())
+	      continue;
+
+	    if(!act.get_dep().end())
+	      root_names.insert(act.get_dep().ParentPkg().Name());
+	  }
+
+	if(root_names.empty())
+	  return "";
+
+	std::string rval;
+
+	bool first = true;
+	for(std::set<std::string>::const_iterator it = root_names.begin();
+	    it != root_names.end(); ++it)
+	  {
+	    if(first)
+	      first = false;
+	    else
+	      rval += ", ";
+
+	    rval += *it;
+	  }
+
+	// ForTranslators: %s is replaced with a comma-delimited list
+	// of package names.
+	return ssprintf(ngettext("(for %s)",
+				 "(for %s)",
+				 root_names.size()),
+			rval.c_str());
+      }
+    else
+      {
+	// If we're being verbose, display whole chains leading to
+	// each target.
+
+	if(reasons.empty())
+	  return "";
+
+	std::string rval;
+	bool first = true;
+	std::sort(reasons.begin(), reasons.end(), compare_first_action());
+	rval += "(";
+	for(std::vector<std::vector<action> >::const_iterator it =
+	      reasons.begin(); it != reasons.end(); ++it)
+	
+	  {
+	    if(it->empty())
+	      continue;
+
+	    if(first)
+	      first = false;
+	    else
+	      rval += ", ";
+
+	    bool first_action = true;
+	    for(std::vector<action>::const_iterator aIt = it->begin();
+		aIt != it->end(); ++aIt)
+	      {
+		if(!first_action)
+		  rval += " ";
+
+		if(!aIt->get_dep().end())
+		  {
+		    const pkgCache::DepIterator &dep(aIt->get_dep());
+
+		    if(first_action)
+		      {
+			rval += const_cast<pkgCache::DepIterator &>(dep).ParentPkg().Name();
+			rval += " ";
+		      }
+
+		    std::string dep_type = const_cast<pkgCache::DepIterator &>(dep).DepType();
+		    rval += cw::util::transcode(cw::util::transcode(dep_type).substr(0, 1));
+		    rval += ": ";
+
+		    rval += const_cast<pkgCache::DepIterator &>(dep).TargetPkg().Name();
+
+		    // Display versioned deps if we're really being
+		    // verbose.
+		    if(verbose > 1 && ((dep->CompareOp & ~pkgCache::Dep::Or) != pkgCache::Dep::NoOp))
+		      {
+			rval += " (";
+			rval += const_cast<pkgCache::DepIterator &>(dep).CompType();
+			rval += " ";
+			rval += dep.TargetVer();
+			rval += ")";
+		      }
+		  }
+		else
+		  {
+		    const pkgCache::PrvIterator &prv(aIt->get_prv());
+		    eassert(!prv.end());
+
+		    if(first_action)
+		      {
+			rval += const_cast<pkgCache::PrvIterator &>(prv).OwnerPkg().Name();
+		      }
+
+		    rval += " ";
+
+		    rval += cw::util::transcode(cw::util::transcode(_("Provides")).substr(0, 1));
+		    rval += "<- ";
+
+		    rval += const_cast<pkgCache::PrvIterator &>(prv).ParentPkg().Name();
+		    if(verbose > 1 && prv.ProvideVersion() != NULL)
+		      {
+			rval += " (";
+			rval += prv.ProvideVersion();
+			rval += ")";
+		      }
+		  }
+
+		first_action = false;
+	      }
+
+	    rval += ")";
+	  }
+
+	return rval;
+      }
+  }
+}
+
 /** Prints a description of a list of packages, with annotations
  *  reflecting how or why they will be changed.
  *
@@ -106,54 +315,117 @@ static string reason_string_list(set<reason> &reasons)
  *  removed, or held.
  *
  *  \param items the set of items to examine
- *  \param showvers if true, display version numbers as appropriate
- *  \param showsize if true, display the change in each package's size
+ *  \param verbose controls various aspects of how verbose the list is.
+ *  \param showvers if \b true, display version numbers as appropriate
+ *  \param showdeps if \b true, display the packages that depend on
+ *                  automatically installed packages.
+ *  \param showsize if \b true, display the change in each package's size
+ *  \param showpurge if \b true, display flags indicating which packages
+ *                   are being purged.
+ *  \param showwhy  if \b true, infer and display the set of manually
+ *                  installed packages that depend on each automatically
+ *                  installed package.
  */
 static void cmdline_show_instinfo(pkgvector &items,
+				  int verbose,
 				  bool showvers,
 				  bool showdeps,
 				  bool showsize,
-				  bool showpurge)
+				  bool showpurge,
+				  bool showwhy)
 {
   sort(items.begin(), items.end(), pkg_byname_compare);
   strvector output;
 
   for(pkgvector::iterator i=items.begin(); i!=items.end(); ++i)
     {
+      std::string tags;
       string s=i->Name();
 
       pkgDepCache::StateCache &state=(*apt_cache_file)[*i];
       //aptitudeDepCache::aptitude_state &extstate=(*apt_cache_file)->get_ext_state(*i);
       pkgCache::VerIterator instver=state.InstVerIter(*apt_cache_file);
 
+      // Set to true if this package should get attached deps.
+      bool deps_ok = showdeps;
+
       if(showpurge)
 	{
 	  if(state.Delete() && state.iFlags&pkgDepCache::Purge)
-	    s += "{p}";
+	    tags.push_back('p');
+	}
+
+      switch(find_pkg_state(*i, *apt_cache_file))
+	{
+	case pkg_auto_remove:
+	case pkg_auto_install:
+	case pkg_auto_hold:
+	  tags.push_back('a');
+	  break;
+	case pkg_unused_remove:
+	  tags.push_back('u');
+	  break;
+	case pkg_broken:
+	  // Do nothing, but don't clear out deps_ok.
+	  break;
+	default:
+	  deps_ok = false;
+	}
+
+      if(!tags.empty())
+	{
+	  std::sort(tags.begin(), tags.end());
+	  s.push_back('{');
+	  s += tags;
+	  s.push_back('}');
 	}
 
       // Display version numbers.
       if(showvers)
 	{
-	  pkgCache::VerIterator cur=i->CurrentVer();
-	  pkgCache::VerIterator cand=state.CandidateVerIter(*apt_cache_file);
+	  pkgCache::VerIterator cur = i->CurrentVer();
+	  pkgCache::VerIterator inst = state.InstVerIter(*apt_cache_file);
 
 	  // Display x -> y for upgraded, held, and downgraded packages.
 	  if( (state.Status==1 || state.Downgrade()) &&
-	      i->CurrentVer()!=cand)
+	      state.Install() &&
+	      i->CurrentVer() != inst)
 	    {
 	      s+=" [";
-	      s+=cur.VerStr();
+	      if(cur.end())
+		s += "??";
+	      else
+		s += cur.VerStr();
 	      s+=" -> ";
-	      s+=cand.VerStr();
+	      if(inst.end())
+		s += "??";
+	      else
+		s += inst.VerStr();
 	      s+="]";
 	    }
-	  else
+	  else if(state.Install())
 	    {
 	      s+=" [";
-	      s+=cand.VerStr();
+	      if(inst.end())
+		s += "??";
+	      else
+		s += inst.VerStr();
 	      s+="]";
 	    }
+	  else if(state.Delete())
+	    {
+	      s += " [";
+	      if(cur.end())
+		{
+		  if((*i)->CurrentState == pkgCache::State::ConfigFiles)
+		    s += _("Config files");
+		  else
+		    s += "??";
+		}
+	      else
+		s += cur.VerStr();
+	      s += "]";
+	    }
 	}
 
       // Show the change in size between the versions.
@@ -175,6 +447,19 @@ static void cmdline_show_instinfo(pkgvector &items,
 	  s+=reason_string_list(reasons);
 	}
 
+      if(showwhy)
+	{
+	  std::string whystring(roots_string(*i, verbose));
+	  if(!whystring.empty())
+	    {
+	      s += " ";
+	      s += whystring;
+	    }
+	}
+
+      if(showvers || showsize || showdeps || showwhy)
+	s += ' ';
+
       output.push_back(s);
     }
 
@@ -265,6 +550,8 @@ static void show_broken_deps(pkgCache::PkgIterator pkg)
     }
 }
 
+// Note that not all of these are actually used, but I'm preserving
+// the tags in case there are requests to reinstate them.
 static const char *cmdline_action_descriptions[num_pkg_action_states]={
   N_("The following packages are BROKEN:"),
   N_("The following packages are unused and will be REMOVED:"),
@@ -463,9 +750,11 @@ static bool prompt_trust()
  */
 static bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 				 pkgset &to_hold, pkgset &to_remove,
-				 bool showvers, bool showdeps, bool showsize,
+				 bool showvers, bool showdeps,
+				 bool showsize, bool showwhy,
 				 int verbose)
 {
+  const int quiet = aptcfg->FindI("Quiet", 0);
   bool all_empty=true;
 
   pkgvector lists[num_pkg_action_states];
@@ -514,14 +803,28 @@ static bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 	  break;
 	}
 
-      if(state!=pkg_unchanged &&
-	 !((state==pkg_auto_hold || state==pkg_hold) && !as_upgrade))
-	lists[state].push_back(pkg);
-
-      if(state==pkg_auto_install)
-	lists[pkg_install].push_back(pkg);
-      else if(state==pkg_auto_remove)
-	lists[pkg_remove].push_back(pkg);
+      switch(state)
+	{
+	case pkg_auto_install:
+	  lists[pkg_install].push_back(pkg);
+	  break;
+	case pkg_unused_remove:
+	case pkg_auto_remove:
+	  lists[pkg_remove].push_back(pkg);
+	  break;
+	case pkg_auto_hold:
+	  if(as_upgrade && to_install.find(pkg) != to_install.end())
+	    lists[pkg_hold].push_back(pkg);
+	  break;
+	case pkg_hold:
+	  if(to_install.find(pkg) != to_install.end())
+	    lists[pkg_hold].push_back(pkg);
+	  break;
+	case pkg_unchanged:
+	  break;
+	default:
+	  lists[state].push_back(pkg);
+	}
     }
 
   for(int i=0; i<num_pkg_action_states; ++i)
@@ -531,29 +834,26 @@ static bool cmdline_show_preview(bool as_upgrade, pkgset &to_install,
 	  all_empty=false;
 
 	  printf("%s\n", _(cmdline_action_descriptions[i]));
-	  if(i==pkg_auto_install || i==pkg_auto_remove || i==pkg_unused_remove ||
-	     i==pkg_auto_hold || i==pkg_broken)
-	    cmdline_show_instinfo(lists[i],
-				  showvers, showdeps, showsize,
-				  (i == pkg_auto_remove ||
-				   i == pkg_unused_remove));
-	  else
-	    cmdline_show_instinfo(lists[i],
-				  showvers, false, showsize,
-				  i == pkg_remove);
+	  cmdline_show_instinfo(lists[i],
+				verbose,
+				showvers, showdeps, showsize,
+				i == pkg_remove ||
+				i == pkg_auto_remove ||
+				i == pkg_unused_remove,
+				showwhy);
 	}
     }
 
-  if(!recommended.empty())
+  if(quiet == 0 && !recommended.empty())
     {
       printf(_("The following packages are RECOMMENDED but will NOT be installed:\n"));
-      cmdline_show_instinfo(recommended, showvers, showdeps, showsize, false);
+      cmdline_show_instinfo(recommended, verbose, showvers, showdeps, showsize, false, showwhy);
     }
 
   if(verbose>0 && !suggested.empty())
     {
       printf(_("The following packages are SUGGESTED but will NOT be installed:\n"));
-      cmdline_show_instinfo(suggested, showvers, showdeps, showsize, false);
+      cmdline_show_instinfo(suggested, verbose, showvers, showdeps, showsize, false, showwhy);
     }
 
   if(all_empty)
@@ -639,7 +939,7 @@ static void cmdline_parse_why(string response)
   splitws(response, arguments, 1, response.size());
 
   if(arguments.empty())
-    printf(_("No packages found -- enter zero or more roots of the search followed by the package to justify."));
+    printf(_("No packages found -- enter zero or more roots of the search followed by the package to justify.\n"));
   else
     {
       bool success;
@@ -658,7 +958,7 @@ static inline cw::fragment *flowindentbox(int i1, int irest, cw::fragment *f)
   return indentbox(i1, irest, flowbox(f));
 }
 
-static void prompt_help(ostream &out)
+static void prompt_help(ostream &out, bool show_resolver_key)
 {
   std::vector<cw::fragment *> fragments;
 
@@ -694,6 +994,11 @@ static void prompt_help(ostream &out)
 				flowindentbox(0, 3,
 					      cw::fragf(_("try to find a reason for installing a single package, or explain why installing one package should lead to installing another package.")))));
 
+  if(show_resolver_key)
+    fragments.push_back(cw::fragf(_("r: %F"),
+				  flowindentbox(0, 3,
+						cw::text_fragment(_("run the automatic dependency resolver to fix the broken dependencies.")))));
+
   fragments.push_back(cw::fragf(_("e: %F"),
 				flowindentbox(0, 3,
 					      cw::fragf(_("enter the full visual interface")))));
@@ -722,6 +1027,8 @@ static void prompt_help(ostream &out)
 				    cw::fragf(_("'&M' to mark packages as automatically installed"))));
   fragments.push_back(flowindentbox(0, 4,
 				    cw::fragf(_("'&m' to mark packages as manually installed"))));
+  fragments.push_back(flowindentbox(0, 4,
+				    cw::fragf(_("'&BD' to install the build-dependencies of a package."))));
 
   cw::fragment *f = indentbox(2, 2, cw::sequence_fragment(fragments));
 
@@ -738,69 +1045,172 @@ bool cmdline_do_prompt(bool as_upgrade,
 		       bool showvers,
 		       bool showdeps,
 		       bool showsize,
+		       bool showwhy,
 		       bool always_prompt,
 		       int verbose,
 		       bool assume_yes,
-		       bool force_no_change)
+		       bool force_no_change,
+		       pkgPolicy &policy,
+		       bool arch_only)
 {
-  bool cont=false;
+  bool exit=false;
   bool rval=true;
   bool first=true;
+  // If true, we will automatically use the internal resolver.  If
+  // false, the internal resolver has failed at least once and so we
+  // should not use it.
+  //
+  // The idea here is that the resolver stays disabled until the
+  // dependencies are resolved (unless explicitly re-enabled), then
+  // becomes available for future breakage.
+  bool use_internal_resolver = true;
 
-  while(!cont)
+  while(!exit)
     {
+      bool have_broken = false;
+      // If we're only doing what the user asked and it's OK to go
+      // ahead, we can break out immediately.
       if(!cmdline_show_preview(true, to_install, to_hold, to_remove,
-			       showvers, showdeps, showsize, verbose) &&
+			       showvers, showdeps, showsize, showwhy, verbose) &&
 	 first &&
 	 !always_prompt &&
 	 (*apt_cache_file)->BrokenCount()==0)
-	cont=true;
+	exit=true;
       else if((*apt_cache_file)->BrokenCount() > 0)
 	{
-	  if(!cmdline_resolve_deps(to_install,
-				   to_hold,
-				   to_remove,
-				   to_purge,
-				   assume_yes,
-				   force_no_change,
-				   verbose))
+	  if(use_internal_resolver)
 	    {
-	      cont=true;
-	      rval=false;
+	      switch(cmdline_resolve_deps(to_install,
+					  to_hold,
+					  to_remove,
+					  to_purge,
+					  assume_yes,
+					  force_no_change,
+					  verbose,
+					  policy,
+					  arch_only))
+		{
+		case aptitude::cmdline::resolver_success:
+		  break;
+		case aptitude::cmdline::resolver_incomplete:
+		  have_broken = true;
+		  use_internal_resolver = false;
+		  break;
+		case aptitude::cmdline::resolver_user_exit:
+		  exit = true;
+		  rval = false;
+		  break;
+		}
+
+	      if(first && assume_yes)
+		{
+		  // If we're supposed to assume "yes", then we actually
+		  // say to abort if there are still broken packages, and
+		  // say to continue otherwise.
+		  rval = !have_broken;
+		  exit = true;
+		}
+
+	      if(!exit)
+		{
+		  // Re-display the preview so the user can see any
+		  // changes the resolver made.
+		  cmdline_show_preview(true, to_install, to_hold, to_remove,
+				       showvers, showdeps, showsize, showwhy,
+				       verbose);
+		}
 	    }
+	  else
+	    // The internal resolver is disabled, fall back to manual
+	    // resolution.
+	    have_broken = true;
 	}
-      else if(assume_yes)
-	cont=true;
-      else
+      else if(first && assume_yes)
+	exit=true;
+
+      // Re-enable the resolver, if it was disabled for some reason
+      // already (e.g., if the dependencies have been fixed manually).
+      if(!have_broken)
+	use_internal_resolver = true;
+
+      if(!exit)
 	{
 	  bool valid_response=false;
 
+	  if(have_broken)
+	    {
+	      if(first)
+		{
+		  const std::string msg = _("aptitude failed to find a solution to these dependencies.  You can solve them yourself by hand or type 'n' to quit.");
+		  cw::fragment *f = cw::text_fragment(msg);
+		  cout << f->layout(screen_width,
+				    screen_width,
+				    cwidget::style());
+		  delete f;
+
+		  show_broken();
+		}
+	    }
+
 	  while(!valid_response)
 	    {
 	      valid_response=true;
 	      fflush(stdout);
 
-	      string response=prompt_string(_("Do you want to continue? [Y/n/?] "));
+	      string prompt =
+		!have_broken
+		    ? _("Do you want to continue? [Y/n/?] ")
+		    : _("Resolve these dependencies by hand? [N/+/-/_/:/?] ");
+
+	      string response=prompt_string(prompt);
 	      string::size_type loc=0;
 
 	      while(loc<response.size() && isspace(response[loc]))
 		++loc;
 
+	      // If the user just pushes Enter, default to accepting
+	      // the current state if there aren't broken packages; if
+	      // there are broken packages, abort.
 	      if(loc==response.size())
 		{
-		  response='y';
+		  response = !have_broken ? 'y' : 'n';
 		  loc=0;
 		}
 
+	      const std::string unknown_key_message =
+		_("Invalid response.  Please enter a valid command or '?' for help.\n");
 	      switch(toupper(response[loc]))
 		{
 		case 'Y':
-		  rval=true;
-		  cont=true;
+		  if(have_broken)
+		    {
+		      cw::fragment *f = flowbox(cw::text_fragment(_("Enter a package management command (such as '+ package' to install a package), 'R' to attempt automatic dependency resolution or 'N' to abort.")));
+		      cout << f->layout(screen_width,
+					screen_width,
+					cwidget::style());
+		      delete f;
+		      
+		      valid_response = false;
+		    }
+		  else
+		    {
+		      rval=true;
+		      exit=true;
+		    }
 		  break;
 		case 'N':
 		  rval=false;
-		  cont=true;
+		  exit=true;
+		  break;
+		case 'R':
+		  if(!have_broken)
+		    {
+		      // Pretend we don't understand.
+		      printf("%s", unknown_key_message.c_str());
+		      valid_response=false;
+		    }
+		  else
+		    use_internal_resolver = true;
 		  break;
 		case 'D':
 		  showdeps=!showdeps;
@@ -839,17 +1249,25 @@ bool cmdline_do_prompt(bool as_upgrade,
 		case '_':
 		case ':':
 		case '&':
-		  cmdline_parse_action(response, to_install, to_hold,
-				       to_remove, to_purge, verbose, true);
+		  {
+		    // Don't play 'which packages have we seen?' games
+		    // now. (should I do two passes like at the
+		    // command-line?)
+		    std::set<pkgCache::PkgIterator> seen_virtual_packages;
+		    cmdline_parse_action(response, seen_virtual_packages,
+					 to_install, to_hold,
+					 to_remove, to_purge, verbose,
+					 policy, arch_only, true);
+		  }
 		  break;
 		case 'E':
 		  ui_preview();
 		case '?':
 		  valid_response=false;
-		  prompt_help(cout);
+		  prompt_help(cout, have_broken);
 		  break;
 		default:
-		  printf(_("Invalid response.  Please enter a valid command or '?' for help.\n"));
+		  printf("%s", unknown_key_message.c_str());
 		  valid_response=false;
 		  break;
 		}
@@ -860,7 +1278,7 @@ bool cmdline_do_prompt(bool as_upgrade,
       if(rval && (!prompt_essential() || !prompt_trust()))
 	{
 	  rval=false;
-	  cont=true;
+	  exit=true;
 	}
 
       first=false;
