@@ -1,4 +1,4 @@
-// cmdline-prompt.h                     -*-c++-*-
+// cmdline_prompt.h                     -*-c++-*-
 //
 //  Copyright 2004 Daniel Burrows
 
@@ -9,6 +9,11 @@
 
 #include <cwidget/generic/util/exception.h>
 
+/** \file cmdline_prompt.h
+ */
+
+class pkgPolicy;
+
 /** Thrown when we get EOF on stdin.  Should never be thrown
  *  to the cwidget::toplevel.
  */
@@ -34,12 +39,18 @@ public:
  *  \param showvers \b true to show version information in the preview.
  *  \param showdeps \b true to show dependency information in the preview.
  *  \param showsize \b true to show size information in the preview.
+ *  \param showwhy  \b true to show the root cause of automatic
+ *                  installs/removals.
  *  \param verbose the current verbosity level
  *  \param assume_yes if \b true, assume the user entered "yes"
  *                    at the prompt.
  *  \param force_no_change if \b true, try extra-hard to preserve
  *                         the user's explicit requests (as
  *                         specified in to_install et al)
+ *  \param policy a policy object, used to look up version priorities.
+ *  \param arch_only if \b true, when the user asks to have build-dependencies
+ *  for a package installed, only the architecture-dependent dependencies
+ *  will be considered.
  *
  *  \throws StdinEOFException
  */
@@ -51,10 +62,13 @@ bool cmdline_do_prompt(bool as_upgrade,
 		       bool showvers,
 		       bool showdeps,
 		       bool showsize,
+		       bool showwhy,
 		       bool always_prompt,
 		       int verbose,
 		       bool assume_yes,
-		       bool force_no_change);
+		       bool force_no_change,
+		       pkgPolicy &policy,
+		       bool arch_only);
 
 /** Prompt for a single line of input from the user.
  *
