@@ -1,6 +1,6 @@
 // cmdline_resolver.cc
 //
-//   Copyright (C) 2005-2007 Daniel Burrows
+//   Copyright (C) 2005-2008 Daniel Burrows
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -41,6 +41,7 @@
 #include <generic/util/util.h>
 
 #include <cwidget/fragment.h>
+#include <cwidget/generic/util/ssprintf.h>
 
 #include <apt-pkg/error.h>
 #include <apt-pkg/strutl.h>
@@ -49,6 +50,8 @@
 #include <fstream>
 #include <sstream>
 
+#include <algorithm>
+
 using namespace std;
 
 typedef generic_solution<aptitude_universe> aptitude_solution;
@@ -78,12 +81,39 @@ static cw::fragment *solution_story(const aptitude_solution &s)
   return cw::sequence_fragment(fragments);
 }
 
+void cmdline_dump_resolver()
+{
+  string dumpfile = aptcfg->Find(PACKAGE "::CmdLine::Resolver-Dump", "");
+  if(!dumpfile.empty())
+    {
+      ofstream f(dumpfile.c_str());
+      if(!f)
+	_error->Errno("dump_resolver", _("Unable to open %s for writing"), dumpfile.c_str());
+      else
+	{
+	  resman->dump(f);
+
+	  if(!f)
+	    _error->Errno("dump_resolver", _("Error writing resolver state to %s"), dumpfile.c_str());
+	  else
+	    cout << _("Resolver state successfully written!");
+	}
+    }
+}
+
 static void setup_resolver(pkgset &to_install,
 			   pkgset &to_hold,
 			   pkgset &to_remove,
 			   pkgset &to_purge,
 			   bool force_no_change)
 {
+  if(!resman->resolver_exists())
+    return;
+
+  // Make sure the resolver is in the initial state so we can twiddle
+  // it.
+  resman->reset_resolver();
+
   resman->set_debug(aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false));
 
   // For all packages that the user listed on the command-line (i.e.,
@@ -118,22 +148,7 @@ static void setup_resolver(pkgset &to_install,
 	}
     }
 
-  string dumpfile = aptcfg->Find(PACKAGE "::CmdLine::Resolver-Dump", "");
-  if(!dumpfile.empty())
-    {
-      ofstream f(dumpfile.c_str());
-      if(!f)
-	_error->Errno("dump_resolver", _("Unable to open %s for writing"), dumpfile.c_str());
-      else
-	{
-	  resman->dump(f);
-
-	  if(!f)
-	    _error->Errno("dump_resolver", _("Error writing resolver state to %s"), dumpfile.c_str());
-	  else
-	    cout << _("Resolver state successfully written!");
-	}
-    }
+  cmdline_dump_resolver();
 }
 
 static inline cw::fragment *flowindentbox(int i1, int irest, cw::fragment *f)
@@ -151,6 +166,7 @@ static void resolver_help(ostream &out)
 				".: %F"
 				"o: %F"
 				"e: %F"
+				"x: %F"
 				"r pkg ver ...: %F%n"
 				"a pkg ver ...: %F%n"
 				"<ACTION> pkg... : %F%n"
@@ -177,9 +193,12 @@ static void resolver_help(ostream &out)
 			      flowindentbox(0, 3,
 					    cw::fragf(_("examine the solution in the visual user interface"))),
 			      flowindentbox(0, 3,
-					    cw::fragf(_("reject the given package versions; don't display any solutions in which they occur."))),
-			      flowindentbox(0, 3,
-					    cw::fragf(_("accept the given package versions; display only solutions in which they occur."))),
+					    cw::fragf(_("abort automatic dependency resolution; resolve dependencies by hand instead"))),
+			      flowindentbox(0, 15,
+					    cw::fragf(_("reject the given package versions; don't display any solutions in which they occur.  Enter UNINST instead of a version to reject removing the package."))),
+			      flowindentbox(0, 15,
+							// ForTranslators: "(;" is ...???? A smiley?
+					    cw::fragf(_("accept the given package versions (; display only solutions in which they occur.  Enter UNINST instead of a version to accept removing the package."))),
 			      flowindentbox(0, 3,
 					    cw::fragf(_("adjust the state of the listed packages, where ACTION is one of:"))),
 			      flowindentbox(0, 4,
@@ -257,7 +276,7 @@ static void reject_or_mandate_version(const string &s,
 
       if(in.eof())
 	{
-	  cerr << ssprintf(_("Expected a version after \"%s\""), pkgname.c_str()) << endl;
+	  cerr << ssprintf(_("Expected a version or \"%s\" after \"%s\""), "UNINST", pkgname.c_str()) << endl;
 	  return;
 	}
 
@@ -468,25 +487,53 @@ public:
   std::string errmsg() const { return msg; }
 };
 
-/** \return the resolver's current solution; if it needs to be calculated
- *          first, run the calculation in the background and display
- *          a spinner in the foreground.
- */
-static
-aptitude_solution calculate_current_solution()
+class CmdlineSearchDisabledException : public cwidget::util::Exception
 {
+  std::string msg;
+
+public:
+  CmdlineSearchDisabledException(const std::string &_msg)
+    : msg(_msg)
+  {
+  }
+
+  std::string errmsg() const { return msg; }
+};
+
+aptitude_solution calculate_current_solution(bool suppress_message)
+{
+  const int step_limit = aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 5000);
+  if(step_limit <= 0)
+    {
+      const std::string msg = ssprintf(_("Would resolve dependencies, but dependency resolution is disabled.\n   (%s::ProblemResolver::StepLimit = 0)\n"), PACKAGE);
+
+      // It's important that the code path leading to here doesn't
+      // access resman: the resolver won't exist in this case.
+      throw CmdlineSearchDisabledException(msg);
+    }
+
+  if(!resman->resolver_exists())
+    {
+      const std::string msg = _("I want to resolve dependencies, but no dependency resolver was created.");
+
+      throw CmdlineSearchAbortedException(msg);
+    }
+
+
+
   if(resman->get_selected_solution() < resman->generated_solution_count())
     return resman->get_solution(resman->get_selected_solution(), 0);
 
 
   cmdline_spinner spin(aptcfg->FindI("Quiet", 0));
 
-  std::cout << "Resolving dependencies..." << std::endl;
+  if(!suppress_message)
+    std::cout << _("Resolving dependencies...") << std::endl;
 
   cwidget::threads::box<cmdline_resolver_continuation::resolver_result> retbox;
 
   resman->get_solution_background(resman->generated_solution_count(),
-				  aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 5000),
+				  step_limit,
 				  new cmdline_resolver_continuation(retbox));
 
   cmdline_resolver_continuation::resolver_result res;
@@ -531,13 +578,16 @@ aptitude_solution calculate_current_solution()
     return res.sol;
 }
 
-bool cmdline_resolve_deps(pkgset &to_install,
-			  pkgset &to_hold,
-			  pkgset &to_remove,
-			  pkgset &to_purge,
-			  bool assume_yes,
-			  bool force_no_change,
-			  int verbose)
+aptitude::cmdline::cmdline_resolver_result
+cmdline_resolve_deps(pkgset &to_install,
+		     pkgset &to_hold,
+		     pkgset &to_remove,
+		     pkgset &to_purge,
+		     bool assume_yes,
+		     bool force_no_change,
+		     int verbose,
+		     pkgPolicy &policy,
+		     bool arch_only)
 {
   bool story_is_default = aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Show-Steps", false);
 
@@ -548,15 +598,15 @@ bool cmdline_resolve_deps(pkgset &to_install,
       aptitude_solution lastsol;
 
       // The inner loop tries to generate solutions until some
-      // packages are modified (then the new set of broken packages,
-      // if any, is displayed and we start over)
+      // packages are modified by the user (then the new set of broken
+      // packages, if any, is displayed and we start over)
       bool modified_pkgs=false;
       while(!modified_pkgs)
 	try
 	  {
 	    try
 	      {
-		aptitude_solution sol = calculate_current_solution();
+		aptitude_solution sol = calculate_current_solution(true);
 
 		if(_error->PendingError())
 		  _error->DumpErrors();
@@ -594,7 +644,7 @@ bool cmdline_resolve_deps(pkgset &to_install,
 		switch(toupper(response[loc]))
 		  {
 		  case 'Y':
-		    (*apt_cache_file)->apply_solution(calculate_current_solution(), NULL);
+		    (*apt_cache_file)->apply_solution(calculate_current_solution(true), NULL);
 		    modified_pkgs=true;
 		    break;
 		  case 'N':
@@ -608,7 +658,10 @@ bool cmdline_resolve_deps(pkgset &to_install,
 		    break;
 		  case 'Q':
 		    cout << _("Abandoning all efforts to resolve these dependencies.") << endl;
-		    return false;
+		    return aptitude::cmdline::resolver_user_exit;
+		  case 'X':
+		    cout << _("Abandoning automatic dependency resolution and reverting to manual resolution.") << endl;
+		    return aptitude::cmdline::resolver_incomplete;
 		  case 'O':
 		    {
 		      story_is_default = !story_is_default;
@@ -645,11 +698,14 @@ bool cmdline_resolve_deps(pkgset &to_install,
 		  case '=':
 		  case '_':
 		  case ':':
-		    /// \todo Maybe only do auto-installation of dependencies
-		    /// after the parse?
-		    cmdline_parse_action(response, to_install, to_hold,
-					 to_remove, to_purge, verbose, false);
-		    modified_pkgs=true;
+		    {
+		      std::set<pkgCache::PkgIterator> seen_virtual_packages;
+		      cmdline_parse_action(response, seen_virtual_packages,
+					   to_install, to_hold,
+					   to_remove, to_purge, verbose,
+					   policy, arch_only, false);
+		      modified_pkgs=true;
+		    }
 		    break;
 		    // Undocumented debug feature:
 		  case '~':
@@ -698,18 +754,18 @@ bool cmdline_resolve_deps(pkgset &to_install,
 		      case 'Y':
 			try
 			  {
-			    calculate_current_solution();
+			    calculate_current_solution(false);
 			    done=true;
 			  }
 			catch(NoMoreTime)
 			  {
 			    // ignore and continue looping.
 			  }
-			// NoMoreExceptions flows to the outer catch.
+			// NoMoreSolutions flows to the outer catch.
 			break;
 		      case 'N':
 			cout << _("Abandoning all efforts to resolve these dependencies.") << endl;
-			return false;
+			return aptitude::cmdline::resolver_incomplete;
 		      default:
 			cout << _("Invalid response; please enter 'y' or 'n'.") << endl;
 		      }
@@ -721,7 +777,7 @@ bool cmdline_resolve_deps(pkgset &to_install,
 	    if(resman->generated_solution_count()==0)
 	      {
 		cout << _("Unable to resolve dependencies!  Giving up...") << endl;
-		return false;
+		return aptitude::cmdline::resolver_incomplete;
 	      }
 	    else
 	      {
@@ -734,6 +790,15 @@ bool cmdline_resolve_deps(pkgset &to_install,
 		lastsol.nullify();
 	      }
 	  }
+	catch(StdinEOFException)
+	  {
+	    throw;
+	  }
+	catch(const CmdlineSearchDisabledException &e)
+	  {
+	    cout << e.errmsg();
+	    return aptitude::cmdline::resolver_incomplete;
+	  }
 	catch(cwidget::util::Exception &e)
 	  {
 	    cout << _("*** ERROR: search aborted by fatal exception.  You may continue\n           searching, but some solutions will be unreachable.")
@@ -742,7 +807,7 @@ bool cmdline_resolve_deps(pkgset &to_install,
 		 << e.errmsg();
 
 	    if(resman->generated_solution_count() == 0)
-	      return false;
+	      return aptitude::cmdline::resolver_incomplete;
 	    else
 	      {
 		cout << endl << endl;
@@ -755,5 +820,189 @@ bool cmdline_resolve_deps(pkgset &to_install,
 	  }
     }
 
-  return true;
+  return aptitude::cmdline::resolver_success;
+}
+
+
+namespace aptitude
+{
+  namespace cmdline
+  {
+    namespace
+    {
+      // Set up aptitude's internal resolver for 'safe' resolution.
+      // Essentialy this means (1) forbid it from removing any package,
+      // and (2) only install the default candidate version of a package
+      // or the current version.
+      void setup_safe_resolver(bool no_new_installs, bool no_new_upgrades)
+      {
+	if(!resman->resolver_exists())
+	  return;
+
+	resman->reset_resolver();
+
+	resman->set_debug(aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false));
+
+	for(pkgCache::PkgIterator p = (*apt_cache_file)->PkgBegin();
+	    !p.end(); ++p)
+	  {
+	    // Forbid the resolver from removing installed packages.
+	    if(!p.CurrentVer().end())
+	      {
+		aptitude_resolver_version
+		  remove_p(p,
+			   pkgCache::VerIterator(*apt_cache_file),
+			   *apt_cache_file);
+
+		resman->reject_version(remove_p);
+	      }
+
+	    // Forbid all real versions that aren't the current version or
+	    // the candidate version.
+	    for(pkgCache::VerIterator v = p.VersionList();
+		!v.end(); ++v)
+	      {
+		// For our purposes, all half-unpacked etc states are
+		// installed.
+		const bool p_is_installed =
+		  p->CurrentState != pkgCache::State::NotInstalled &&
+		  p->CurrentState != pkgCache::State::ConfigFiles;
+
+		const bool p_will_install =
+		  (*apt_cache_file)[p].Install();
+
+		if(v != p.CurrentVer() &&
+		   // Disallow installing not-installed packages that
+		   // aren't marked for installation if
+		   // no_new_installs is set.
+		   ((!p_is_installed && !p_will_install && no_new_installs) ||
+		    (p_is_installed && p_will_install && no_new_upgrades) ||
+		    v != (*apt_cache_file)[p].CandidateVerIter(*apt_cache_file)))
+		  {
+		    aptitude_resolver_version
+		      p_v(p, v, *apt_cache_file);
+		    resman->reject_version(p_v);
+		  }
+	      }
+	  }
+
+	cmdline_dump_resolver();
+      }
+    }
+
+    // Take the first solution we can compute, returning false if we
+    // failed to find a solution.
+    bool safe_resolve_deps(int verbose, bool no_new_installs, bool no_new_upgrades)
+    {
+      if(!resman->resolver_exists())
+	return true;
+
+      const bool debug = aptcfg->FindB(PACKAGE "::CmdLine::Resolver-Debug", false);
+
+      setup_safe_resolver(no_new_installs, no_new_upgrades);
+
+      generic_solution<aptitude_universe> last_sol;
+      try
+	{
+	  generic_solution<aptitude_universe> sol = calculate_current_solution(true);
+	  last_sol = sol;
+	  bool first = true;
+	  typedef imm::map<aptitude_resolver_package,
+	    generic_solution<aptitude_universe>::action> actions_map;
+	  while(!(!first &&
+		  last_sol &&
+		  sol.get_actions() == last_sol.get_actions() &&
+		  sol.get_unresolved_soft_deps() == last_sol.get_unresolved_soft_deps()))
+	    {
+	      first = false;
+	      // Mandate all the upgrades and installs in this solution,
+	      // then ask for the next solution.  The goal is to try to
+	      // find the best solution, not just some solution, but to
+	      // preserve our previous decisions in order to avoid
+	      // thrashing around between alternatives.
+	      for(actions_map::const_iterator it = sol.get_actions().begin();
+		  it != sol.get_actions().end(); ++it)
+		{
+		  if(it->second.ver.get_ver() != it->second.ver.get_pkg().CurrentVer())
+		    {
+		      if(debug)
+			std::cout << "Mandating the upgrade or install " << it->second.ver
+				  << ", since it was produced as part of a solution."
+				  << std::endl;
+		      resman->mandate_version(it->second.ver);
+		    }
+		  else
+		    {
+		      if(debug)
+			std::cout << "Not mandating " << it->second.ver
+				  << ", since it is a hold."
+				  << std::endl;
+		    }
+		}
+	      last_sol = sol;
+	      resman->select_next_solution();
+	      sol = calculate_current_solution(false);
+	    }
+
+	  // Internal error that should never happen, but try to survive
+	  // if it does.
+	  std::cout << "*** Internal error: the resolver unexpectedly produced the same result twice."
+		    << std::endl
+		    << "Previous iteration was:" << std::endl;
+	  last_sol.dump(std::cout, true);
+	  std::cout << std::endl << "Current iteration was:" << std::endl;
+	  sol.dump(std::cout, true);
+	  std::cout << std::endl;
+
+	  (*apt_cache_file)->apply_solution(last_sol, NULL);
+	}
+      // If anything goes wrong, we give up (silently if verbosity is disabled).
+      catch(NoMoreTime)
+	{
+	  if(last_sol)
+	    {
+	      std::cout << _("The resolver timed out after producing a solution; some possible upgrades might not be performed.");
+	      (*apt_cache_file)->apply_solution(last_sol, NULL);
+	      return true;
+	    }
+	  else
+	    {
+	      if(verbose > 0)
+		std::cout << _("Unable to resolve dependencies for the upgrade (the resolver timed out).");
+	      return false;
+	    }
+	}
+      catch(NoMoreSolutions)
+	{
+	  if(last_sol)
+	    {
+	      (*apt_cache_file)->apply_solution(last_sol, NULL);
+	      return true;
+	    }
+	  else
+	    {
+	      if(verbose > 0)
+		std::cout << _("Unable to resolve dependencies for the upgrade (no solution found).");
+	      return false;
+	    }
+	}
+      catch(const cw::util::Exception &e)
+	{
+	  if(last_sol)
+	    {
+	      (*apt_cache_file)->apply_solution(last_sol, NULL);
+	      std::cout << cw::util::ssprintf(_("Dependency resolution incomplete (%s); some possible upgrades might not be performed."), e.errmsg().c_str());
+	      return true;
+	    }
+	  else
+	    {
+	      if(verbose > 0)
+		std::cout << cw::util::ssprintf(_("Unable to resolve dependencies for the upgrade (%s)."), e.errmsg().c_str());
+	      return false;
+	    }
+	}
+
+      return true;
+    }
+  }
 }
