@@ -5,11 +5,22 @@
 #ifndef CMDLINE_UPGRADE_H
 #define CMDLINE_UPGRADE_H
 
+#include "cmdline_util.h"
+
+#include <vector>
+
+/** \file cmdline_upgrade.h
+ */
+
 int cmdline_upgrade(int argc, char *argv[],
 		    const char *status_fname, bool simulate,
+		    bool no_new_installs,
 		    bool assume_yes, bool download_only,
-		    bool showvers, bool showdeps, bool showsize,
+		    bool showvers, bool showdeps,
+		    bool showsize, bool showwhy,
+		    const std::vector<aptitude::cmdline::tag_application> &user_tags,
 		    bool visual_preview, bool always_prompt,
-		    bool queue_only, int verbose);
+		    bool arch_only, bool queue_only,
+		    int verbose);
 
 #endif // CMDLINE_UPGRADE_H
