@@ -1,6 +1,6 @@
 // cmdline_why.cc                                -*-c++-*-
 //
-//   Copyright (C) 2007 Daniel Burrows
+//   Copyright (C) 2007-2008 Daniel Burrows
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -22,6 +22,7 @@
 
 #include "cmdline_common.h"
 #include "cmdline_show.h"
+#include "cmdline_util.h"
 
 #include <aptitude.h>
 
@@ -53,6 +54,8 @@ namespace cw = cwidget;
 using cwidget::fragf;
 using cwidget::fragment;
 
+using namespace aptitude::matching;
+
 namespace
 {
   struct compare_pkgs
@@ -63,33 +66,76 @@ namespace
       return p1->ID < p2->ID;
     }
   };
+}
 
-  // Represents a single step in a justification.  This means that
-  // we either follow a dependency or a Provides.
-  //
-  // NB: the only reason for tagging with the id is because I don't
-  // want massive copying, and I'm lazy, so I want to reuse immsets
-  // instead of writing a proper singly linked, refcounted list type.
-  class justify_action
+namespace aptitude
+{
+  namespace why
   {
-    pkgCache::DepIterator dep;
-    pkgCache::PrvIterator prv;
-    int id;
-
-  public:
-    justify_action(const pkgCache::DepIterator &_dep,
-	   int _id)
-      : dep(_dep), id(_id)
+    // Represents a full forward or reverse justification for the
+    // installation of the given package/version.  The justification
+    // may terminate on either a package version, a provided package
+    // name, or the removal of a package.
+    class justification
     {
-    }
+      target the_target;
+      imm::set<action> actions;
 
-    justify_action(const pkgCache::PrvIterator &_prv,
-		   int _id)
-      : prv(_prv), id(_id)
-    {
-    }
+      justification(const target &_the_target,
+		    const imm::set<action> &_actions)
+	: the_target(_the_target), actions(_actions)
+      {
+      }
+    public:
+      // Create a node with an empty history rooted at the given target.
+      justification(const target &_the_target)
+	: the_target(_the_target)
+      {
+      }
 
-    cw::style get_style() const
+      const target &get_target() const
+      {
+	return the_target;
+      }
+
+      const imm::set<action> &get_actions() const
+      {
+	return actions;
+      }
+
+      // Generate all the successors of this node (Q: could I lift
+      // justify_target::generate_successors into this class?  It feels
+      // like it makes more sense to have the smarts in here)
+      void generate_successors(std::deque<justification> &output,
+			       const search_params &params,
+			       int verbosity) const
+      {
+	the_target.generate_successors(*this, output, params, verbosity);
+      }
+
+      // Build a successor of this node.
+      justification successor(const target &new_target,
+			      const pkgCache::DepIterator &dep) const
+      {
+	imm::set<action> new_actions(actions);
+	new_actions.insert(action(dep, new_actions.size()));
+
+	return justification(new_target, new_actions);
+      }
+
+      justification successor(const target &new_target,
+			      const pkgCache::PrvIterator &prv) const
+      {
+	imm::set<action> new_actions(actions);
+	new_actions.insert(action(prv, new_actions.size()));
+
+	return justification(new_target, new_actions);
+      }
+
+      cwidget::fragment *description() const;
+    };
+
+    cw::style action::get_style() const
     {
       pkgCache::PkgIterator pkg;
       if(!dep.end())
@@ -100,7 +146,7 @@ namespace
       return pkg_item::pkg_style(pkg, false);
     }
 
-    cw::fragment *description_column1_fragment() const
+    cw::fragment *action::description_column1_fragment() const
     {
       pkgCache::VerIterator ver;
       if(!dep.end())
@@ -138,7 +184,7 @@ namespace
       return cw::text_fragment(rval);
     }
 
-    cw::fragment *description_column2_fragment() const
+    cw::fragment *action::description_column2_fragment() const
     {
       if(!dep.end())
 	return cw::text_fragment(const_cast<pkgCache::DepIterator &>(dep).DepType());
@@ -146,7 +192,7 @@ namespace
 	return cw::text_fragment(_("Provides"));
     }
 
-    cw::fragment *description_column3_fragment() const
+    cw::fragment *action::description_column3_fragment() const
     {
       // Q: can I use a std::string and cw::util::transcode on the way out instead?
       if(!dep.end())
@@ -159,15 +205,15 @@ namespace
 	return cw::text_fragment(const_cast<pkgCache::PrvIterator &>(prv).ParentPkg().Name());
     }
 
-    cw::fragment *description_fragment() const
+    cw::fragment *action::description_fragment() const
     {
       return cw::fragf("%F %F %F",
-		   description_column1_fragment(),
-		   description_column2_fragment(),
-		   description_column3_fragment());
+		       description_column1_fragment(),
+		       description_column2_fragment(),
+		       description_column3_fragment());
     }
 
-    bool operator<(const justify_action &other) const
+    bool action::operator<(const action &other) const
     {
       typedef pkgCache::Dependency Dependency;
       typedef pkgCache::Provides Provides;
@@ -199,68 +245,8 @@ namespace
       else
 	return false;
     }
-  };
 
-  class justify_node;
-
-  class search_params
-  {
-  public:
-    enum VersionSelection { Current, Candidate, Install };
-    enum DepLevel { DependsOnly, Recommends, Suggests };
-
-  private:
-    VersionSelection version_selection;
-    DepLevel dep_level;
-    bool allow_choices;
-
-  public:
-    search_params(VersionSelection _version_selection,
-		  DepLevel _dep_level,
-		  bool _allow_choices)
-      : version_selection(_version_selection),
-	dep_level(_dep_level),
-	allow_choices(_allow_choices)
-    {
-    }
-
-    bool should_follow_dep(const pkgCache::DepIterator &dep) const
-    {
-      switch(dep->Type)
-	{
-	case pkgCache::Dep::Depends:
-	case pkgCache::Dep::PreDepends:
-	case pkgCache::Dep::Conflicts:
-	case pkgCache::Dep::DpkgBreaks:
-	  return true;
-	case pkgCache::Dep::Recommends:
-	  return dep_level == Recommends || dep_level == Suggests;
-	case pkgCache::Dep::Suggests:
-	  return dep_level == Suggests;
-	default:
-	  return false;
-	}
-    }
-
-    pkgCache::VerIterator selected_version(const pkgCache::PkgIterator &pkg) const
-    {
-      switch(version_selection)
-	{
-	case Current:
-	  return pkg.CurrentVer();
-	case Candidate:
-	  return (*apt_cache_file)[pkg].CandidateVerIter(*apt_cache_file);
-	case Install:
-	  return (*apt_cache_file)[pkg].InstVerIter(*apt_cache_file);
-	default:
-	  fprintf(stderr, "Unknown version selection, something is very wrong.\n");
-	  return pkg.CurrentVer();
-	}
-    }
-
-    bool get_allow_choices() const { return allow_choices; }
-
-    std::wstring description() const
+    std::wstring search_params::description() const
     {
       std::wstring rval(L"{ ");
       rval += W_("dep_level");
@@ -296,6 +282,9 @@ namespace
 	case Install:
 	  rval += W_("Install");
 	  break;
+	case InstallNotCurrent:
+	  rval += W_("InstallNotCurrent");
+	  break;
 	default:
 	  rval += L"???";
 	  break;
@@ -312,206 +301,13 @@ namespace
       rval += L" }";
       return rval;
     }
-  };
 
-  /** Represents something that can be justified.  This is the tail of
-   *  a justification deduction and also used to input targets.
-   */
-  class justify_target
-  {
-    // The three possible states are:
-    //
-    // pkg.end() and ver.end() are false
-    //      => this is a specific package version.
-    // pkg.end() is false, ver.end() is true, is_provided_name is true
-    //      => this is a provided package name.
-    // pkg.end() is false, ver.end() is true, is_provided_name is false
-    //      => this is a package removal.
-
-    // An Install node represents the installation of a package.
-    //
-    // A Remove node represents the removal of a package.
-    //
-    // A ProvidesInstall node indicates that something that provides
-    // the package name is being installed.
-    //
-    // A ProvidesRemove node indicates that something that provides
-    // the package name is being removed.
-    enum NodeType { InstallType, RemoveType, ProvidesInstall, ProvidesRemove };
-
-    NodeType node_type;
-    // The package on which to operate; since we uniformly take one of
-    // the three package targets, there isn't a need to store a
-    // version.
-    pkgCache::PkgIterator pkg;
-    // The provides is stored so we can use the ProvideVersion.
-    pkgCache::PrvIterator prv;
-
-    justify_target(const pkgCache::PkgIterator &_pkg,
-		   const pkgCache::PrvIterator &_prv,
-		   NodeType _node_type)
-      : pkg(_pkg), prv(_prv), node_type(_node_type)
-    {
-    }
-
-    justify_target(const pkgCache::PkgIterator &_pkg,
-		   NodeType _node_type)
-      : pkg(_pkg), node_type(_node_type)
-    {
-    }
-
-    static justify_target Provide(const pkgCache::PkgIterator &pkg,
-				  const pkgCache::PrvIterator &prv,
-				  bool provided_is_removed)
-    {
-      NodeType tp = provided_is_removed ? ProvidesRemove : ProvidesInstall;
-      return justify_target(pkg, prv, tp);
-    }
-
-  public:
-    /** Returns the package that should be marked as "visited"
-     *  to restrict future searches. (since the search never visits
-     *  two different versions of the same package, this is OK)
-     */
-    pkgCache::PkgIterator get_visited_package() const
-    {
-      return pkg;
-    }
-
-    pkgCache::PrvIterator get_provides() const
-    {
-      return prv;
-    }
-
-    /** \brief Return true if this is a ProvidesInstall or ProvidesRemove node. */
-    bool is_provides() const
-    {
-      return node_type == ProvidesInstall || node_type == ProvidesRemove;
-    }
-
-    /** \brief Return true if this is a Remove or ProvidesRemove node. */
-    bool is_remove() const
-    {
-      return node_type == RemoveType || node_type == ProvidesRemove;
-    }
-
-    // Use the implicit operator==.
-    //
-    // bool operator==(const justify_target &other) const;
-
-    static justify_target Install(const pkgCache::PkgIterator &pkg)
-    {
-      return justify_target(pkg, InstallType);
-    }
-
-    static justify_target Remove(const pkgCache::PkgIterator &pkg)
-    {
-      return justify_target(pkg, RemoveType);
-    }
-
-    cw::fragment *description() const
-    {
-      pkgCache::PkgIterator &mpkg = const_cast<pkgCache::PkgIterator &>(pkg);
-
-      switch(node_type)
-	{
-	case InstallType:
-	  return cw::fragf(_("Install(%s)"), mpkg.Name());
-	case RemoveType:
-	  return cw::fragf(_("Remove(%s)"), mpkg.Name());
-	case ProvidesInstall:
-	  return cw::fragf(_("Install(%s provides %s)"),
-		       const_cast<pkgCache::PrvIterator &>(prv).OwnerPkg().Name(),
-		       mpkg.Name());
-	case ProvidesRemove:
-	  return cw::fragf(_("Remove(%s provides %s)"),
-		       const_cast<pkgCache::PrvIterator &>(prv).OwnerPkg().Name(),
-		       mpkg.Name());
-	}
-    }
-
-    /** \brief Push the successors of this target onto the head of a queue.
-     *
-     *  The successors of a node are the actions that could have generated
-     *  it (installing this package, installing this provides).
-     *
-     *  \param parent the parent of any new search nodes that are generated.
-     *  \param output the queue onto which the successors should be loaded.
-     *  \param params the parameters of the search (these control
-     *                which dependencies get followed).
-     */
-    void generate_successors(const justify_node &parent,
-			     std::deque<justify_node> &output,
-			     const search_params &params,
-			     int verbosity) const;
-  };
-
-  // Represents a full forward or reverse justification for the
-  // installation of the given package/version.  The justification may
-  // terminate on either a package version, a provided package name,
-  // or the removal of a package.
-  class justify_node
-  {
-    justify_target target;
-    imm::set<justify_action> actions;
-
-    justify_node(const justify_target &_target,
-		 const imm::set<justify_action> &_actions)
-      : target(_target), actions(_actions)
-    {
-    }
-  public:
-    // Create a node with an empty history rooted at the given target.
-    justify_node(const justify_target &_target)
-      : target(_target)
-    {
-    }
-
-    const justify_target &get_target() const
-    {
-      return target;
-    }
-
-    const imm::set<justify_action> &get_actions() const
-    {
-      return actions;
-    }
-
-    // Generate all the successors of this node (Q: could I lift
-    // justify_target::generate_successors into this class?  It feels
-    // like it makes more sense to have the smarts in here)
-    void generate_successors(std::deque<justify_node> &output,
-			     const search_params &params,
-			     int verbosity) const
-    {
-      target.generate_successors(*this, output, params, verbosity);
-    }
-
-    // Build a successor of this node.
-    justify_node successor(const justify_target &target,
-			   const pkgCache::DepIterator &dep) const
-    {
-      imm::set<justify_action> new_actions(actions);
-      new_actions.insert(justify_action(dep, new_actions.size()));
-
-      return justify_node(target, new_actions);
-    }
-
-    justify_node successor(const justify_target &target,
-			   const pkgCache::PrvIterator &prv) const
-    {
-      imm::set<justify_action> new_actions(actions);
-      new_actions.insert(justify_action(prv, new_actions.size()));
-
-      return justify_node(target, new_actions);
-    }
-
-    cw::fragment *description() const
+    cw::fragment *justification::description() const
     {
       std::vector<cw::fragment *> rval;
-      rval.push_back(cw::fragf("%F\n", target.description()));
+      rval.push_back(cw::fragf("%F\n", the_target.description()));
       std::vector<cw::fragment *> col1_entries, col2_entries, col3_entries;
-      for(imm::set<justify_action>::const_iterator it = actions.begin();
+      for(imm::set<action>::const_iterator it = actions.begin();
 	  it != actions.end(); ++it)
 	{
 	  col1_entries.push_back(cw::hardwrapbox(cw::fragf("%F | \n", it->description_column1_fragment())));
@@ -524,13 +320,12 @@ namespace
       std::vector<fragment_column_entry> columns;
       columns.push_back(fragment_column_entry(false, true, 0, fragment_column_entry::top, col1_entries));
       columns.push_back(fragment_column_entry(false, false, 1, fragment_column_entry::top, NULL)),
-      columns.push_back(fragment_column_entry(false, true, 0, fragment_column_entry::top, col2_entries));
+	columns.push_back(fragment_column_entry(false, true, 0, fragment_column_entry::top, col2_entries));
       columns.push_back(fragment_column_entry(false, false, 1, fragment_column_entry::top, NULL)),
-      columns.push_back(fragment_column_entry(false, true, 0, fragment_column_entry::top, col3_entries));
+	columns.push_back(fragment_column_entry(false, true, 0, fragment_column_entry::top, col3_entries));
       rval.push_back(fragment_columns(columns));
       return cw::sequence_fragment(rval);
     }
-  };
 
   cw::fragment *print_dep(pkgCache::DepIterator dep)
   {
@@ -548,11 +343,31 @@ namespace
 		   dep.TargetPkg().Name());
   }
 
+  cw::fragment *target::description() const
+  {
+    pkgCache::PkgIterator &mpkg = const_cast<pkgCache::PkgIterator &>(pkg);
 
-  void justify_target::generate_successors(const justify_node &parent,
-					   std::deque<justify_node> &output,
-					   const search_params &params,
-					   int verbosity) const
+    switch(type)
+      {
+      case InstallType:
+	return cw::fragf(_("Install(%s)"), mpkg.Name());
+      case RemoveType:
+	return cw::fragf(_("Remove(%s)"), mpkg.Name());
+      case ProvidesInstall:
+	return cw::fragf(_("Install(%s provides %s)"),
+			 const_cast<pkgCache::PrvIterator &>(prv).OwnerPkg().Name(),
+			 mpkg.Name());
+      case ProvidesRemove:
+	return cw::fragf(_("Remove(%s provides %s)"),
+			 const_cast<pkgCache::PrvIterator &>(prv).OwnerPkg().Name(),
+			 mpkg.Name());
+      }
+  }
+
+  void target::generate_successors(const justification &parent,
+				   std::deque<justification> &output,
+				   const search_params &params,
+				   int verbosity) const
   {
     // The reverse successors of an install node are all the revdeps
     // of the package, minus conflicts and deps from versions that
@@ -638,6 +453,47 @@ namespace
 	    continue;
 	  }
 
+	if(params.get_only_not_current())
+	  {
+	    bool satisfied_by_current = false;
+
+	    // Skip this dep if it's satisfied by the package's
+	    // current version.
+	    if(is_provides())
+	      {
+		pkgCache::VerIterator provider_current = get_provides().OwnerPkg().CurrentVer();
+		for(pkgCache::PrvIterator prv = provider_current.ProvidesList();
+		    !satisfied_by_current && !prv.end(); ++prv)
+		  {
+		    if(dep.TargetVer() == NULL ||
+		       (prv.ProvideVersion() != NULL &&
+			_system->VS->CheckDep(prv.ProvideVersion(),
+					      dep->CompareOp,
+					      dep.TargetVer())))
+		      satisfied_by_current = true;
+		  }
+	      }
+	    else
+	      {
+		if((*apt_cache_file)[dep.TargetPkg()].Status != 2)
+		  {
+		    pkgCache::VerIterator current = dep.TargetPkg().CurrentVer();
+		    if(dep.TargetVer() == NULL ||
+		       _system->VS->CheckDep(current.VerStr(),
+					     dep->CompareOp,
+					     dep.TargetVer()))
+		      satisfied_by_current = true;
+		  }
+	      }
+
+	    if(satisfied_by_current)
+	      {
+		if(verbosity > 1)
+		  std::cout << _("    ++   --> skipping, the dep is satisfied by the current version\n");
+		continue;
+	      }
+	  }
+
 	const char *ver_to_check;
 	if(is_provides())
 	  ver_to_check = get_provides().ProvideVersion();
@@ -664,13 +520,13 @@ namespace
 	  {
 	    if(verbosity > 1)
 	      std::cout << _("    ++   --> ENQUEUING\n");
-	    justify_target target(Install(dep.ParentPkg()));
-	    output.push_back(parent.successor(target, dep));
+	    target the_target(Install(dep.ParentPkg()));
+	    output.push_back(parent.successor(the_target, dep));
 	  }
 	else
 	  {
 	    if(verbosity > 1)
-	      std::cout << _("    ++   --> skipping, version check failed.\n");
+	      std::cout << _("    ++   --> skipping, version check failed\n");
 	  }
       }
 
@@ -691,19 +547,20 @@ namespace
 		  std::cout << ssprintf(_("    ++   --> ENQUEUING %s Provides %s\n"),
 					prv.OwnerPkg().Name(),
 					prv.ParentPkg().Name());
-		justify_target target(Provide(prv.ParentPkg(), prv, is_remove()));
-		output.push_back(parent.successor(target, prv));
+		target the_target(Provide(prv.ParentPkg(), prv, is_remove()));
+		output.push_back(parent.successor(the_target, prv));
 	      }
 	  }
       }
   }
 
-
+    namespace
+    {
   class justification_search
   {
     // The central queue.  Nodes are inserted at the back and removed
     // from the front.
-    std::deque<justify_node> q;
+    std::deque<justification> q;
 
     std::vector<pkg_matcher *> leaves;
 
@@ -737,8 +594,7 @@ namespace
      *                suggests/recommends to be important.
      */
     justification_search(const std::vector<pkg_matcher *> &_leaves,
-			 const pkgCache::PkgIterator &root,
-			 bool search_for_removal,
+			 const target &root,
 			 const search_params &_params,
 			 int _verbosity)
       : leaves(_leaves),
@@ -748,10 +604,7 @@ namespace
 	verbosity(_verbosity)
     {
       // Prime the pump.
-      if(search_for_removal)
-	q.push_back(justify_node(justify_target::Remove(root)));
-      else
-	q.push_back(justify_node(justify_target::Install(root)));
+      q.push_back(justification(root));
     }
 
     justification_search(const justification_search &other)
@@ -803,9 +656,9 @@ namespace
      *
      *  \return true if a justification was found, false otherwise.
      */
-    bool next(std::vector<justify_action> &output)
+    bool next(std::vector<action> &output)
     {
-      std::vector<justify_action> tmp;
+      std::vector<action> tmp;
       bool reached_leaf = false;
 
       if(seen_packages == NULL)
@@ -828,7 +681,7 @@ namespace
 	{
 	  // NB: could avoid this copy, but I'd have to move the
 	  // pop_front() into all execution branches.  Not worth it.
-	  const justify_node front(q.front());
+	  const justification front(q.front());
 	  q.pop_front();
 
 
@@ -849,18 +702,28 @@ namespace
 	  bool &package_is_seen = seen_packages[frontid];
 	  if(package_is_seen)
 	    continue;
-	  package_is_seen = true;
+	  // Don't flag the starting package as "seen", since we want
+	  // to be able to find self-loops.
+	  if(!front.get_actions().empty())
+	    package_is_seen = true;
 
-	  // Test whether the front node is a leaf; if it is, return
-	  // it and quit.
+	  // If we've stepped at least once, test whether the front
+	  // node is a leaf; if it is, return it and quit.
+	  //
+	  // Checking that we stepped at least once (verified by
+	  // checking whether we have any actions in our result list)
+	  // ensures that we always return nontrivial answers (i.e.,
+	  // even if the target of the search matches a leaf pattern,
+	  // we'll keep looking past it).
 	  pkgCache::VerIterator frontver = params.selected_version(frontpkg);
-	  if(!frontver.end())
+	  if(!frontver.end() && !front.get_actions().empty())
 	    for(std::vector<pkg_matcher *>::const_iterator it = leaves.begin();
 		!reached_leaf && it != leaves.end(); ++it)
 	      {
-		if((*it)->matches(frontpkg, frontver,
-				  *apt_cache_file,
-				  *apt_package_records))
+		if(apply_matcher((*it),
+				 frontpkg, frontver,
+				 *apt_cache_file,
+				 *apt_package_records))
 		  reached_leaf = true;
 	      }
 	  if(reached_leaf)
@@ -879,49 +742,36 @@ namespace
       return reached_leaf;
     }
   };
-
-  // Silly matcher that only matches one package.
-  class const_matcher : public pkg_matcher
-  {
-    pkgCache::PkgIterator match_pkg;
-
-    class const_name_result : public pkg_match_result
-    {
-      std::string name_group;
-    public:
-      const_name_result(const std::string &_name_group)
-	: name_group(_name_group)
-      {
-      }
-
-      unsigned int num_groups() { return 1; }
-      const std::string &group(unsigned int n) { return name_group; }
-    };
-  public:
-    const_matcher(const pkgCache::PkgIterator &_match_pkg)
-      : match_pkg(_match_pkg)
-    {
     }
 
-    bool matches(const pkgCache::PkgIterator &pkg,
-		 const pkgCache::VerIterator &ver,
-		 aptitudeDepCache &cache,
-		 pkgRecords &records)
+    bool find_justification(const target &target,
+			    const std::vector<aptitude::matching::pkg_matcher *> leaves,
+			    const search_params &params,
+			    bool find_all,
+			    std::vector<std::vector<action> > &output)
     {
-      return pkg == match_pkg;
-    }
+      justification_search search(leaves, target, params, 0);
 
-    pkg_match_result *get_match(const pkgCache::PkgIterator &pkg,
-				const pkgCache::VerIterator &ver,
-				aptitudeDepCache &cache,
-				pkgRecords &records)
-    {
-      if(pkg == match_pkg)
-	return new const_name_result(pkg.Name());
+      std::vector<std::vector<action> > rval;
+      std::vector<action> tmp;
+
+      int i = 0;
+
+      while((i == 0 || find_all) && search.next(tmp))
+	{
+	  rval.push_back(std::vector<action>());
+	  rval.back().swap(tmp);
+	}
+
+      if(rval.size() > 0)
+	{
+	  output.swap(rval);
+	  return true;
+	}
       else
-	return NULL;
+	return false;
     }
-  };
+  }
 }
 
 cw::fragment *do_why(const std::vector<pkg_matcher *> &leaves,
@@ -930,6 +780,8 @@ cw::fragment *do_why(const std::vector<pkg_matcher *> &leaves,
 		 bool root_is_removal,
 		 bool &success)
 {
+  using namespace aptitude::why;
+
   std::vector<cw::fragment *> rval;
   success = true;
 
@@ -1025,14 +877,15 @@ cw::fragment *do_why(const std::vector<pkg_matcher *> &leaves,
   // might not perfectly eliminate results that appear identical if
   // multiple versions of something are available; needs more work to
   // do that)
-  std::set<std::vector<justify_action> > seen_results;
-  std::vector<justify_action> results;
+  std::set<std::vector<action> > seen_results;
+  std::vector<action> results;
   bool first = true;
 
   for(std::vector<search_params>::const_iterator it = searches.begin();
       it != searches.end(); ++it)
     {
-      justification_search search(leaves, root, root_is_removal, *it, verbosity);
+      target goal = root_is_removal ? target::Remove(root) : target::Install(root);
+      justification_search search(leaves, goal, *it, verbosity);
 
       while(search.next(results))
 	{
@@ -1045,18 +898,22 @@ cw::fragment *do_why(const std::vector<pkg_matcher *> &leaves,
 	  else
 	    seen_results.insert(results);
 
-	  if(first)
-	    first = false;
-	  else
-	    rval.push_back(cw::newline_fragment());
-
 	  if(results.empty())
-	    return cw::fragf(_("The package \"%s\" is a starting point of the search.\n"),
-			 root.Name());
+	    // This is a starting point of the search, but just keep
+	    // going and try to find other people who depend on it.
+	    // NB: this shouldn't be necessary (the search itself
+	    // should filter out empty results) but it's left in for
+	    // the sake of defensive programming.
+	    continue;
 	  else
 	    {
+	      if(first)
+		first = false;
+	      else
+		rval.push_back(cw::newline_fragment());
+
 	      std::vector<cw::fragment *> col1_entries, col2_entries, col3_entries;
-	      for(std::vector<justify_action>::const_iterator it = results.begin();
+	      for(std::vector<action>::const_iterator it = results.begin();
 		  it != results.end(); ++it)
 		{
 		  col1_entries.push_back(cw::hardwrapbox(cw::style_fragment(cw::fragf("%F\n", it->description_column1_fragment()),
@@ -1110,9 +967,9 @@ cw::fragment *do_why(const std::vector<pkg_matcher *> &leaves,
       success = false;
 
       if(root_is_removal)
-	return cw::fragf(_("No justification for removing %s could be constructed.\n"), root.Name());
+	return cw::fragf(_("Unable to find a reason to remove %s.\n"), root.Name());
       else
-	return cw::fragf(_("No justification for %s could be constructed.\n"), root.Name());
+	return cw::fragf(_("Unable to find a reason to install %s.\n"), root.Name());
     }
   else
     return cw::sequence_fragment(rval);
@@ -1153,13 +1010,13 @@ bool interpret_why_args(const std::vector<std::string> &args,
     {
       // If there isn't a tilde, treat it as an exact package name.
       pkg_matcher *m = NULL;
-      if(it->find('~') == std::string::npos)
+      if(!cmdline_is_search_pattern(*it))
 	{
 	  pkgCache::PkgIterator pkg = (*apt_cache_file)->FindPkg(*it);
 	  if(pkg.end())
 	    _error->Error(_("No package named \"%s\" exists."), it->c_str());
 	  else
-	    m = new const_matcher(pkg);
+	    m = make_const_matcher(pkg);
 	}
       else
 	m = parse_pattern(*it);
@@ -1262,17 +1119,7 @@ int cmdline_why(int argc, char *argv[],
       if(m == NULL)
 	parsing_arguments_failed = true;
       else
-	{
-	  matchers.push_back(m);
-	  if(!pkg.end() && m->matches(pkg, pkg.CurrentVer(),
-				      *apt_cache_file, *apt_package_records))
-	    {
-	      printf(_("The package \"%s\" is manually installed.\n"),
-		     pkg.Name());
-	      delete m;
-	      return 0;
-	    }
-	}
+	matchers.push_back(m);
     }
 
 
