@@ -1,6 +1,6 @@
 // apt.h  -*-c++-*-
 //
-//  Copyright 1999-2002, 2004-2005, 2007 Daniel Burrows
+//  Copyright 1999-2002, 2004-2005, 2007-2008 Daniel Burrows
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -17,10 +17,7 @@
 //  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 //  Boston, MA 02111-1307, USA.
 //
-//  Argh.  It seems that it's pretty much necessary to have a central
-// repository for apt structures -- the cache in particular.  This is it.
-// No class wrapper because, well, the cache is already a class and you can't
-// open more than one at once, so there :)
+
 
 #ifndef APT_H
 #define APT_H
@@ -31,6 +28,17 @@
 
 #include <utility>
 
+/** \brief Central repository for apt structures
+ *
+ * 
+ *  Argh.  It seems that it's pretty much necessary to have a central
+ *  repository for apt structures -- the cache in particular.  This is it.
+ *  No class wrapper because, well, the cache is already a class and you can't
+ *  open more than one at once, so there :)
+ * 
+ *  \file apt.h
+ */
+
 class OpProgress;
 class pkgRecords;
 class pkgSourceList;
@@ -74,6 +82,13 @@ void apt_init(OpProgress *progess_bar,
 /** Close the cache file and destroy the associated data structures. */
 void apt_close_cache();
 
+/** \return \b true if the version of the apt library against
+ *  which we are running supports the RootDir convention.
+ *
+ *  This returns \b false if apt_preinit() has not yet been invoked.
+ */
+bool get_apt_knows_about_rootdir();
+
 /** If the cache is closed, open it; otherwise do nothing.
  *
  *  \param progress_bar a progress bar with which to display the
@@ -264,4 +279,18 @@ inline bool is_conflict(pkgCache::Dep::DepType type)
   return is_conflict(static_cast<unsigned char>(type));
 }
 
+namespace aptitude
+{
+  namespace apt
+  {
+    /** \return \b true if the given dependency is a Replaces dependency
+     *  and participates in a conflicts/provides/replaces relationship.
+     *
+     *  Note that replaces and conflicts relationships without a
+     *  version aren't considered; should they be?
+     */
+    bool is_full_replacement(const pkgCache::DepIterator &dep);
+  }
+}
+
 #endif
