@@ -1,6 +1,6 @@
 // main.cc  (ne testscr.cc)
 //
-//  Copyright 1999-2007 Daniel Burrows
+//  Copyright 1999-2008 Daniel Burrows
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -34,6 +34,7 @@
 
 #include <generic/apt/apt.h>
 #include <generic/apt/config_signal.h>
+#include <generic/apt/matchers.h>
 
 #include <generic/problemresolver/exceptions.h>
 
@@ -49,6 +50,7 @@
 #include <cmdline/cmdline_do_action.h>
 #include <cmdline/cmdline_download.h>
 #include <cmdline/cmdline_dump_resolver.h>
+#include <cmdline/cmdline_extract_cache_subset.h>
 #include <cmdline/cmdline_forget_new.h>
 #include <cmdline/cmdline_moo.h>
 #include <cmdline/cmdline_prompt.h>
@@ -56,6 +58,7 @@
 #include <cmdline/cmdline_show.h>
 #include <cmdline/cmdline_update.h>
 #include <cmdline/cmdline_upgrade.h>
+#include <cmdline/cmdline_user_tag.h>
 #include <cmdline/cmdline_why.h>
 
 #include <sigc++/functors/ptr_fun.h>
@@ -105,12 +108,25 @@ static void show_version()
 #ifdef __GNUC__
   printf(_("Compiler: g++ %s\n"), __VERSION__);
 #endif
-#ifdef NCURSES_VERSION
-  printf(_("\nNCurses version: %s\n"), NCURSES_VERSION);
+  printf("%s", _("Compiled against:\n"));
+  printf(_("  apt version %d.%d.%d\n"),
+	 APT_PKG_MAJOR, APT_PKG_MINOR, APT_PKG_RELEASE);
+#ifndef NCURSES_VERSION
+  printf(_("  NCurses version: Unknown\n"));
 #else
-  printf(_("\nCurses version: %s\n"), curses_version());
+  printf(_("  NCurses version %s\n"), NCURSES_VERSION);
 #endif
-  printf(_("libsigc++ version: %s\n"), SIGC_VERSION);
+  printf(_("  libsigc++ version: %s\n"), SIGC_VERSION);
+#ifdef HAVE_EPT
+  printf(_("  Ept support enabled.\n"));
+#else
+  printf(_("  Ept support disabled.\n"));
+#endif
+
+  printf("%s", _("\nCurrent library versions:\n"));
+  printf(_("  NCurses version: %s\n"), curses_version());
+  printf(_("  cwidget version: %s\n"), cwidget::version().c_str());
+  printf(_("  Apt version: %s\n"), pkgLibVersion);
 }
 
 static void usage()
@@ -128,10 +144,6 @@ static void usage()
   printf(_(" unhold       - Cancel a hold command for a package\n"));
   printf(_(" markauto     - Mark packages as having been automatically installed\n"));
   printf(_(" unmarkauto   - Mark packages as having been manually installed\n"));
-  // This is out of line with the rest to avoid messing up
-  // translations.  Eventually all the usage information will move to
-  // a table-layout-based approach, which will avoid this sort of
-  // problem.
   printf(_(" forbid-version - Forbid aptitude from upgrading to a specific package version.\n"));
   printf(_(" update       - Download lists of new/upgradable packages\n"));
   printf(_(" safe-upgrade - Perform a safe upgrade\n"));
@@ -144,6 +156,11 @@ static void usage()
   printf(_(" changelog    - View a package's changelog\n"));
   printf(_(" download     - Download the .deb file for a package\n"));
   printf(_(" reinstall    - Download and (possibly) reinstall a currently installed package\n"));
+  printf(_(" why          - Show the manually installed packages that require a package, or\n"
+           "                why one or more packages would require the given package\n"));
+  printf(_(" why-not      - Show the manually installed packages that lead to a conflict\n"
+           "                with the given package, or why one or more packages would\n"
+           "                lead to a conflict with the given package if installed\n"));
   printf("\n");
   printf(_("  Options:\n"));
   printf(_(" -h             This help text\n"));
@@ -160,9 +177,11 @@ static void usage()
   printf(_(" -Z             Show the change in installed size of each package.\n"));
   printf(_(" -v             Display extra information. (may be supplied multiple times)\n"));
   printf(_(" -t [release]   Set the release from which packages should be installed\n"));
-  printf(_(" -q             In command-line mode, suppress the incremental progress indicators.\n"));
+  printf(_(" -q             In command-line mode, suppress the incremental progress\n"
+           "                indicators.\n"));
   printf(_(" -o key=val     Directly set the configuration option named 'key'\n"));
-  printf(_(" --with(out)-recommends	Specify whether or not to treat recommends as\n                strong dependencies\n"));
+  printf(_(" --with(out)-recommends	Specify whether or not to treat recommends as\n"
+           "                strong dependencies\n"));
   printf(_(" -S fname       Read the aptitude extended status info from fname.\n"));
   printf(_(" -u             Download new package lists on startup.\n"));
   printf(_(" -i             Perform an install run on startup.\n"));
@@ -176,6 +195,19 @@ enum {
   OPTION_VISUAL_PREVIEW,
   OPTION_QUEUE_ONLY,
   OPTION_PURGE_UNUSED,
+  OPTION_ALLOW_UNTRUSTED,
+  OPTION_NO_NEW_INSTALLS,
+  OPTION_NO_NEW_UPGRADES,
+  OPTION_ALLOW_NEW_INSTALLS,
+  OPTION_ALLOW_NEW_UPGRADES,
+  OPTION_ADD_USER_TAG,
+  OPTION_ADD_USER_TAG_TO,
+  OPTION_REMOVE_USER_TAG,
+  OPTION_REMOVE_USER_TAG_FROM,
+  OPTION_SAFE_RESOLVER,
+  OPTION_FULL_RESOLVER,
+  OPTION_ARCH_ONLY,
+  OPTION_NOT_ARCH_ONLY
 };
 int getopt_result;
 
@@ -186,6 +218,7 @@ option opts[]={
   {"quiet", 2, NULL, 'q'},
   {"width", 1, NULL, 'w'},
   {"simulate", 0, NULL, 's'},
+  {"allow-untrusted", 0, &getopt_result, OPTION_ALLOW_UNTRUSTED},
   {"with-recommends", 0, NULL, 'r'},
   {"without-recommends", 0, NULL, 'R'},
   {"download-only", 0, NULL, 'd'},
@@ -193,12 +226,25 @@ option opts[]={
   {"verbose", 0, NULL, 'v'},
   {"show-versions", 0, NULL, 'V'},
   {"show-deps", 0, NULL, 'D'},
+  {"show-why", 0, NULL, 'W'},
   {"prompt", 0, NULL, 'P'},
   {"sort", 1, NULL, 'O'},
   {"target-release", 1, NULL, 't'},
+  {"no-new-installs", 0, &getopt_result, OPTION_NO_NEW_INSTALLS},
+  {"no-new-upgrades", 0, &getopt_result, OPTION_NO_NEW_UPGRADES},
+  {"allow-new-installs", 0, &getopt_result, OPTION_ALLOW_NEW_INSTALLS},
+  {"allow-new-upgrades", 0, &getopt_result, OPTION_ALLOW_NEW_UPGRADES},
+  {"safe-resolver", 0, &getopt_result, OPTION_SAFE_RESOLVER},
+  {"full-resolver", 0, &getopt_result, OPTION_FULL_RESOLVER},
   {"visual-preview", 0, &getopt_result, OPTION_VISUAL_PREVIEW},
   {"schedule-only", 0, &getopt_result, OPTION_QUEUE_ONLY},
   {"purge-unused", 0, &getopt_result, OPTION_PURGE_UNUSED},
+  {"add-user-tag", 1, &getopt_result, OPTION_ADD_USER_TAG},
+  {"add-user-tag-to", 1, &getopt_result, OPTION_ADD_USER_TAG_TO},
+  {"remove-user-tag", 1, &getopt_result, OPTION_REMOVE_USER_TAG},
+  {"remove-user-tag-from", 1, &getopt_result, OPTION_REMOVE_USER_TAG_FROM},
+  {"arch-only", 0, &getopt_result, OPTION_ARCH_ONLY},
+  {"not-arch-only", 0, &getopt_result, OPTION_NOT_ARCH_ONLY},
   {0,0,0,0}
 };
 
@@ -227,18 +273,35 @@ int main(int argc, char *argv[])
   bool simulate = aptcfg->FindB(PACKAGE "::CmdLine::Simulate", false) ||
     aptcfg->FindB(PACKAGE "::Simulate", false);
   bool download_only=aptcfg->FindB(PACKAGE "::CmdLine::Download-Only", false);;
+  bool arch_only = aptcfg->FindB("Apt::Get::Arch-Only", false);
 
   bool update_only=false, install_only=false, queue_only=false;
   bool assume_yes=aptcfg->FindB(PACKAGE "::CmdLine::Assume-Yes", false);
   bool fix_broken=aptcfg->FindB(PACKAGE "::CmdLine::Fix-Broken", false);
+  bool safe_upgrade_no_new_installs = aptcfg->FindB(PACKAGE "::CmdLine::Safe-Upgrade::No-New-Installs", false);
+  bool safe_resolver_no_new_installs = aptcfg->FindB(PACKAGE "::Safe-Resolver::No-New-Installs", false);
+  bool safe_resolver_no_new_upgrades = aptcfg->FindB(PACKAGE "::Safe-Resolver::No-New-Upgrades", false);
+  bool always_use_safe_resolver = aptcfg->FindB(PACKAGE "::Always-Use-Safe-Resolver", false);
+  bool safe_resolver_option = false;
+  bool full_resolver_option = false;
+
+  // This tracks whether we got a --*-new-installs command-line
+  // argument, so we can present a useful warning message to the user
+  // if it's not applicable.
+  bool saw_new_installs_option = false;
+  // Same for --*-new-upgrades
+  bool saw_new_upgrades_option = false;
+
   bool showvers=aptcfg->FindB(PACKAGE "::CmdLine::Show-Versions", false);
   bool showdeps=aptcfg->FindB(PACKAGE "::CmdLine::Show-Deps", false);
   bool showsize=aptcfg->FindB(PACKAGE "::CmdLine::Show-Size-Changes", false);
+  bool showwhy = aptcfg->FindB(PACKAGE "::CmdLine::Show-Why", false);
   bool visual_preview=aptcfg->FindB(PACKAGE "::CmdLine::Visual-Preview", false);
   bool always_prompt=aptcfg->FindB(PACKAGE "::CmdLine::Always-Prompt", false);
   int verbose=aptcfg->FindI(PACKAGE "::CmdLine::Verbose", 0);
   bool seen_quiet = false;
   int quiet = 0;
+  std::vector<aptitude::cmdline::tag_application> user_tags;
 
   int curopt;
   // The last option seen
@@ -259,7 +322,7 @@ int main(int argc, char *argv[])
     aptcfg->Set(PACKAGE "::Delete-Unused-Pattern", "");
 
   // Read the arguments:
-  while((curopt=getopt_long(argc, argv, "DVZvhS:uiF:w:sO:fdyPt:q::Rro:", opts, NULL))!=-1)
+  while((curopt=getopt_long(argc, argv, "DVZWvhS:uiF:w:sO:fdyPt:q::Rro:", opts, NULL))!=-1)
     {
       switch(curopt)
 	{
@@ -323,10 +386,10 @@ int main(int argc, char *argv[])
 	  seen_quiet = true;
 	  break;
 	case 'r':
-	  aptcfg->SetNoUser(PACKAGE "::Recommends-Important", "true");
+	  aptcfg->SetNoUser("Apt::Install-Recommends", "true");
 	  break;
 	case 'R':
-	  aptcfg->SetNoUser(PACKAGE "::Recommends-Important", "false");
+	  aptcfg->SetNoUser("Apt::Install-Recommends", "false");
 	  aptcfg->SetNoUser(PACKAGE "::Keep-Recommends", "true");
 	  break;
 	case 't':
@@ -339,7 +402,7 @@ int main(int argc, char *argv[])
 
 	    // note that if eqloc!=s.npos, s.size()>0
 	    if(eqloc==s.npos || eqloc==s.size()-1)
-	      fprintf(stderr, _("-o requires an argument of the form key=value, got %s"), optarg);
+	      fprintf(stderr, _("-o requires an argument of the form key=value, got %s\n"), optarg);
 	    else
 	      {
 		string key(s, 0, eqloc), value(s, eqloc+1);
@@ -360,6 +423,9 @@ int main(int argc, char *argv[])
 	case 'Z': // ew
 	  showsize=true;
 	  break;
+	case 'W':
+	  showwhy = true;
+	  break;
 	case 'P':
 	  always_prompt=true;
 	  break;
@@ -373,6 +439,35 @@ int main(int argc, char *argv[])
 	    case OPTION_VERSION:
 	      show_version();
 	      exit(0);
+	    case OPTION_ALLOW_UNTRUSTED:
+	      aptcfg->Set(PACKAGE "::CmdLine::Ignore-Trust-Violations", true);
+	      break;
+	    case OPTION_NO_NEW_INSTALLS:
+	      safe_upgrade_no_new_installs = true;
+	      safe_resolver_no_new_installs = true;
+	      saw_new_installs_option = true;
+	      break;
+	    case OPTION_ALLOW_NEW_INSTALLS:
+	      safe_upgrade_no_new_installs = false;
+	      safe_resolver_no_new_installs = false;
+	      saw_new_installs_option = true;
+	      break;
+	    case OPTION_NO_NEW_UPGRADES:
+	      safe_resolver_no_new_upgrades = true;
+	      saw_new_upgrades_option = false;
+	      break;
+	    case OPTION_ALLOW_NEW_UPGRADES:
+	      safe_resolver_no_new_upgrades = false;
+	      saw_new_upgrades_option = false;
+	      break;
+	    case OPTION_SAFE_RESOLVER:
+	      always_use_safe_resolver = true;
+	      safe_resolver_option = true;
+	      break;
+	    case OPTION_FULL_RESOLVER:
+	      always_use_safe_resolver = false;
+	      full_resolver_option = true;
+	      break;
 	    case OPTION_VISUAL_PREVIEW:
 	      visual_preview=true;	      
 	      break;
@@ -382,6 +477,49 @@ int main(int argc, char *argv[])
 	    case OPTION_PURGE_UNUSED:
 	      aptcfg->Set(PACKAGE "::Purge-Unused", "true");
 	      break;
+	    case OPTION_ADD_USER_TAG:
+	    case OPTION_REMOVE_USER_TAG:
+	      {
+		using aptitude::cmdline::tag_application;
+		const bool is_add = (getopt_result == OPTION_ADD_USER_TAG);
+		user_tags.push_back(tag_application(is_add, optarg, NULL));
+		break;
+	      }
+	    case OPTION_ADD_USER_TAG_TO:
+	    case OPTION_REMOVE_USER_TAG_FROM:
+	      {
+		using aptitude::cmdline::tag_application;
+		const bool is_add = (getopt_result == OPTION_ADD_USER_TAG);
+		const std::string arg = optarg;
+		const std::string::size_type splitloc = arg.find(',');
+		if(splitloc == arg.npos)
+		  {
+		    fprintf(stderr, _("No comma following tag name \"%s\".\n"), arg.c_str());
+		    return -1;
+		  }
+		else
+		  {
+		    const std::string patternstr(arg, splitloc + 1);
+		    const std::vector<const char *> terminators;
+		    aptitude::matching::pkg_matcher *m =
+		      aptitude::matching::parse_pattern(patternstr,
+							terminators);
+
+		    if(m == NULL)
+		      {
+			_error->DumpErrors();
+			return -1;
+		      }
+
+		    user_tags.push_back(tag_application(is_add, optarg, m));
+		  }
+	      }
+	    case OPTION_NOT_ARCH_ONLY:
+	      arch_only = false;
+	      break;
+	    case OPTION_ARCH_ONLY:
+	      arch_only = true;
+	      break;
 	    default:
 	      fprintf(stderr, "%s",
 		      _("WEIRDNESS: unknown option code received\n"));
@@ -398,10 +536,11 @@ int main(int argc, char *argv[])
 	}
     }
 
-  if(quiet == 0 && !isatty(1))
-    aptcfg->SetNoUser("quiet", 1);
-  else if(seen_quiet)
+  int curr_quiet = aptcfg->FindI("quiet", 0);
+  if(seen_quiet)
     aptcfg->SetNoUser("quiet", quiet);
+  if(quiet == 0 && !isatty(1))
+    aptcfg->SetNoUser("quiet", std::max(curr_quiet, 1));
 
   if(simulate)
     aptcfg->SetNoUser(PACKAGE "::Simulate", true);
@@ -417,7 +556,7 @@ int main(int argc, char *argv[])
 
   if((update_only || install_only) && optind!=argc)
     {
-      fprintf(stderr, "%s",
+      fprintf(stderr, "%s\n",
 	      _("-u and -i may not be specified in command-line mode (eg, with 'install')"));
       usage();
       exit(1);
@@ -438,14 +577,17 @@ int main(int argc, char *argv[])
 
 	  if(update_only || install_only)
 	    {
-	      fprintf(stderr, "%s",
+	      fprintf(stderr, "%s\n",
 		      _("-u and -i may not be specified with a command"));
 	      usage();
 	      exit(1);
 	    }
 
+	  // TODO: warn the user if they passed --full-resolver to
+	  // something other than "upgrade" or do_action.
+
 	  if(!strcasecmp(argv[optind], "update"))
-	    return cmdline_update(argc-optind, argv+optind);
+	    return cmdline_update(argc-optind, argv+optind, verbose);
 	  else if(!strcasecmp(argv[optind], "clean"))
 	    return cmdline_clean(argc-optind, argv+optind, simulate);
 	  else if(!strcasecmp(argv[optind], "autoclean"))
@@ -476,21 +618,48 @@ int main(int argc, char *argv[])
 		   (!strcasecmp(argv[optind], "unmarkauto")) ||
 		   (!strcasecmp(argv[optind], "forbid-version")) ||
 		   (!strcasecmp(argv[optind], "keep")) ||
-		   (!strcasecmp(argv[optind], "keep-all")) )
-	    return cmdline_do_action(argc-optind, argv+optind,
-				     status_fname,
-				     simulate, assume_yes, download_only,
-				     fix_broken, showvers, showdeps, showsize,
-				     visual_preview, always_prompt,
-				     queue_only, verbose);
+		   (!strcasecmp(argv[optind], "keep-all")) ||
+		   (!strcasecmp(argv[optind], "build-dep")) ||
+		   (!strcasecmp(argv[optind], "build-depends")))
+	    {
+	      return cmdline_do_action(argc-optind, argv+optind,
+				       status_fname,
+				       simulate, assume_yes, download_only,
+				       fix_broken, showvers, showdeps,
+				       showsize, showwhy,
+				       visual_preview, always_prompt,
+				       always_use_safe_resolver,
+				       safe_resolver_no_new_installs, safe_resolver_no_new_upgrades,
+				       user_tags,
+				       arch_only, queue_only, verbose);
+	    }
 	  else if(!strcasecmp(argv[optind], "safe-upgrade") ||
 		  !strcasecmp(argv[optind], "upgrade"))
-	    return cmdline_upgrade(argc-optind, argv+optind,
-				   status_fname,
-				   simulate, assume_yes, download_only,
-				   showvers, showdeps, showsize,
-				   visual_preview, always_prompt,
-				   queue_only, verbose);
+	    {
+	      if(full_resolver_option || safe_resolver_option)
+		{
+		  fprintf(stderr, _("The options --safe-resolver and --full-resolver are not meaningful\nfor %s, which always uses the safe resolver.\n"),
+			  argv[optind]);
+		  return -1;
+		}
+
+	      return cmdline_upgrade(argc-optind, argv+optind,
+				     status_fname, simulate,
+				     safe_upgrade_no_new_installs,
+				     assume_yes, download_only,
+				     showvers, showdeps,
+				     showsize, showwhy,
+				     user_tags,
+				     visual_preview, always_prompt,
+				     arch_only, queue_only, verbose);
+	    }
+	  else if(!strcasecmp(argv[optind], "add-user-tag") ||
+		  !strcasecmp(argv[optind], "remove-user-tag"))
+	    return aptitude::cmdline::cmdline_user_tag(argc - optind, argv + optind,
+						       quiet, verbose);
+	  else if(!strcasecmp(argv[optind], "extract-cache-subset"))
+	    return aptitude::cmdline::extract_cache_subset(argc - optind,
+							   argv + optind);
 	  else if(!strcasecmp(argv[optind], "download"))
 	    return cmdline_download(argc-optind, argv+optind);
 	  else if(!strcasecmp(argv[optind], "changelog"))
@@ -541,7 +710,7 @@ int main(int argc, char *argv[])
 
 	  std::string backtrace = e.get_backtrace();
 	  if(!backtrace.empty())
-	    fprintf(stderr, _("Backtrace:\n%s"), backtrace.c_str());
+	    fprintf(stderr, _("Backtrace:\n%s\n"), backtrace.c_str());
 	  return -1;
 	}
     }
@@ -590,7 +759,7 @@ int main(int argc, char *argv[])
 
       std::string backtrace = e.get_backtrace();
       if(!backtrace.empty())
-	fprintf(stderr, _("Backtrace:\n%s"), backtrace.c_str());
+	fprintf(stderr, _("Backtrace:\n%s\n"), backtrace.c_str());
       return -1;
     }
 
