@@ -670,7 +670,7 @@ void pkg_item::pkg_columnizer::setup_columns(bool force_update)
       if(!cw::util::transcode(aptcfg->Find(PACKAGE "::UI::Package-Display-Format",
 				 default_pkgdisplay).c_str(),
 		    cfg))
-	_error->Errno("iconv", _("Unable to cw::util::transcode package display format after \"%ls\""), cfg.c_str());
+	_error->Errno("iconv", _("Unable to transcode package display format after \"%ls\""), cfg.c_str());
       else
 	columns=parse_columns(cfg,
 			      pkg_columnizer::parse_column_type,
@@ -679,7 +679,7 @@ void pkg_item::pkg_columnizer::setup_columns(bool force_update)
 	{
 	  cfg.clear();
 	  if(!cw::util::transcode(default_pkgdisplay, cfg))
-	    _error->Errno("iconv", _("Unable to cw::util::transcode package display format after \"%ls\""), cfg.c_str());
+	    _error->Errno("iconv", _("Unable to transcode package display format after \"%ls\""), cfg.c_str());
 	  else
 	    columns=parse_columns(cfg,
 				  pkg_columnizer::parse_column_type,
