@@ -1,6 +1,6 @@
 // pkg_grouppolicy.cc
 //
-//  Copyright 1999-2005, 2007 Daniel Burrows
+//  Copyright 1999-2005, 2007-2008 Daniel Burrows
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -28,6 +28,7 @@
 #include <cwidget/generic/util/transcode.h>
 
 #include <generic/apt/apt.h>
+#include <generic/apt/config_signal.h>
 #include <generic/apt/matchers.h>
 #include <generic/apt/pkg_hier.h>
 #include <generic/apt/tags.h>
@@ -40,6 +41,7 @@
 
 #include <ctype.h>
 
+#include <apt-pkg/configuration.h>
 #include <apt-pkg/pkgsystem.h>
 #include <apt-pkg/version.h>
 
@@ -47,10 +49,11 @@
 #include <sigc++/trackable.h>
 
 using namespace std;
-using namespace __gnu_cxx;
 
 namespace cw = cwidget;
 
+namespace match = aptitude::matching;
+
 pkg_grouppolicy_factory::~pkg_grouppolicy_factory()
 {
 }
@@ -94,6 +97,7 @@ public:
   virtual void add_package(const pkgCache::PkgIterator &i, pkg_subtree *root)
     {
       root->add_child(new pkg_item(i, get_sig()));
+      root->inc_num_packages();
     }
 };
 
@@ -110,20 +114,20 @@ pkg_grouppolicy *pkg_grouppolicy_end_factory::instantiate(pkg_signal *_sig,
 
 class pkg_grouppolicy_section:public pkg_grouppolicy
 {
-  typedef hash_map<string, pair<pkg_grouppolicy *, pkg_subtree *> > section_map;
+  typedef std::map<string, pair<pkg_grouppolicy *, pkg_subtree *> > section_map;
   section_map sections;
 
   pkg_grouppolicy_factory *chain;
 
   // As in the factory
-  int split_mode;
+  pkg_grouppolicy_section_factory::split_mode_type split_mode;
   bool passthrough;
 
   // The descriptions are in the cw::style used by package descriptions.
-  static hash_map<string, wstring> section_descriptions;
+  static std::map<string, wstring> section_descriptions;
   static void init_section_descriptions();
 public:
-  pkg_grouppolicy_section(int _split_mode,
+  pkg_grouppolicy_section(pkg_grouppolicy_section_factory::split_mode_type _split_mode,
 			  bool _passthrough,
 			  pkg_grouppolicy_factory *_chain,
 			  pkg_signal *_sig,
@@ -150,144 +154,234 @@ pkg_grouppolicy *pkg_grouppolicy_section_factory::instantiate(pkg_signal *_sig,
 				     _sig, _desc_sig);
 }
 
-hash_map<string, wstring> pkg_grouppolicy_section::section_descriptions;
+std::map<string, wstring> pkg_grouppolicy_section::section_descriptions;
+
+namespace
+{
+  std::string unquote_section_description(const std::string &desc)
+  {
+    std::string rval;
+    // Replace "\\n" with a newline and "''" with a double-quote.
+    // The latter replacement is necessary because apt doesn't know
+    // about backslash-escaped double quotes (Bug #260446).
+    std::string::size_type pos = 0;
+
+    while(pos != desc.size())
+      {
+	switch(desc[pos])
+	  {
+	  case '\'':
+	    if(pos + 1 < desc.size() &&
+	       desc[pos + 1] == '\'')
+	      {
+		rval += '"';
+		pos += 2;
+	      }
+	    else
+	      {
+		rval += '\'';
+		++pos;
+	      }
+	    break;
+	  case '\\':
+	    if(pos + 1 < desc.size())
+	      {
+		switch(desc[pos + 1])
+		  {
+		  case 'n':
+		    rval += '\n';
+		    break;
+		  case 't':
+		    rval += '\t';
+		    break;
+		  default:
+		    rval += desc[pos + 1];
+		    break;
+		  }
+
+		pos += 2;
+	      }
+	    else
+	      {
+		rval += '\\';
+		++pos;
+	      }
+	    break;
+	  default:
+	    rval += desc[pos];
+	    ++pos;
+	    break;
+	  }
+      }
+    return rval;
+  }
+}
 
-// Should this be externally configurable?
 void pkg_grouppolicy_section::init_section_descriptions()
 {
   static bool already_done=false;
 
+  // TODO: this is wrong -- we need to reset already_done when
+  // anything in the Sections hierarchy changes!
   if(already_done)
     return;
 
-  section_descriptions["Tasks"]=W_("Packages which set up your computer to perform a particular task\n Packages in the 'Tasks' section contain no files; they merely depend upon other packages. These packages provide an easy way to select a predefined set of packages for a specialized task.");
+  // Read in the top-level section descriptions.
+  const Configuration::Item * const Root = aptcfg->Tree(PACKAGE "::Sections::Descriptions");
+  if(Root != NULL)
+    {
+      for(const Configuration::Item *Curr = Root->Child;
+	  Curr != NULL; Curr = Curr->Next)
+	{
+	  const std::string desc = Curr->Value;
+	  const std::string final_desc = unquote_section_description(desc);;
 
-  section_descriptions["Unknown"]=W_("Packages with no declared section\n No section is given for these packages. Perhaps there is an error in the Packages file?");
+	  section_descriptions[Curr->Tag] = cw::util::transcode(final_desc);
+	}
+    }
 
-  section_descriptions["admin"]=W_("Administrative utilities (install software, manage users, etc)\n Packages in the 'admin' section allow you to perform administrative tasks such as installing software, managing users, configuring and monitoring your system, examining network traffic, and so on.");
-
-  section_descriptions["alien"]=W_("Packages converted from foreign formats (rpm, tgz, etc)\n Packages in the 'alien' section were created by the 'alien' program from a non-Debian package format such as RPM");
-  section_descriptions["base"]=W_("The Debian base system\n Packages in the 'base' section are part of the initial system installation.");
-  section_descriptions["comm"]=W_("Programs for faxmodems and other communication devices\n Packages in the 'comm' section are used to control modems and other hardware communications devices. This includes software to control faxmodems (for instance, PPP for dial-up internet connections and programs originally written for that purpose, such as zmodem/kermit), as well as software to control cellular phones, interface with FidoNet, and run a BBS.");
-  section_descriptions["devel"]=W_("Utilities and programs for software development\n Packages in the 'devel' section are used to write new software and work on existing software. Non-programmers who do not compile their own software probably do not need much software from this section.\n .\n It includes compilers, debugging tools, programmer's editors, source processing tools, and other things related to software development.");
-  section_descriptions["doc"]=W_("Documentation and specialized programs for viewing documentation\n Packages in the 'doc' section document parts of the Debian system, or are viewers for documentation formats.");
-  section_descriptions["editors"]=W_("Text editors and word processors\n Packages in the 'editors' section allow you to edit plain ASCII text. These are not necessarily word processors, although some word processors may be found in this section.");
-  section_descriptions["electronics"]=W_("Programs for working with circuits and electronics\n Packages in the 'electronics' section include circuit design tools, simulators and assemblers for microcontrollers, and other related software.");
-  section_descriptions["embedded"]=W_("Programs for embedded systems\n Packages in the 'embedded' section are meant to run on embedded devices. Embedded devices are specialized hardware devices with much less power than a typical desktop system: for instance, a PDA, a cell phone, or a Tivo.");
-  section_descriptions["gnome"]=W_("The GNOME Desktop System\n GNOME is a collection of software which provides an easy-to-use desktop environment for Linux.  Packages in the 'gnome' section are part of the GNOME environment or closely integrated into it.");
-  section_descriptions["games"]=W_("Games, toys, and fun programs\n Packages in the 'games' section are meant primarily for entertainment.");
-  section_descriptions["graphics"]=W_("Utilities to create, view, and edit graphics files\n Packages in the 'graphics' section include viewers for image files, image processing and manipulation software, software to interact with graphics hardware (such as video cards, scanners, and digital cameras), and programming tools for handling graphics.");
-  section_descriptions["hamradio"]=W_("Software for ham radio operators\n Packages in the 'hamradio' section are meant primarily for ham radio operators.");
-  section_descriptions["interpreters"]=W_("Interpreters for interpreted languages\n Packages in the 'interpreters' section include interpreters for languages like Python, Perl, and Ruby, and libraries for these same languages.");
-  section_descriptions["kde"]=W_("The KDE Desktop System\n KDE is a collection of software which provides an easy-to-use desktop environment for Linux.  Packages in the 'kde' section are part of the KDE environment or closely integrated into it.");
-  section_descriptions["libdevel"]=W_("Development files for libraries\n Packages in the 'libdevel' section contain files required for building programs that use libraries in the 'libs' section.  You don't need packages from this section unless you want to compile software yourself.");
-  section_descriptions["libs"]=W_("Collections of software routines\n Packages in the 'libs' section provide necessary functionality for other software on the computer. With very few exceptions, you should not need to explicitly install a package from this section; the package system will install them as required to fulfill dependencies.");
-  section_descriptions["perl"]=W_("Perl interpreter and libraries\n Packages in the 'perl' section provide the Perl programming language and many third-party libraries for it. Unless you are a Perl programmer, you don't need to install packages from this section explicitly; the package system will install them if they are required.");
-  section_descriptions["python"]=W_("Python interpreter and libraries\n Packages in the 'python' section provide the Python programming language and many third-party libraries for it. Unless you are a Python programmer, you don't need to install packages from this section explicitly; the package system will install them if they are required.");
-  section_descriptions["mail"]=W_("Programs to write, send, and route email messages\n Packages in the 'mail' section include mail readers, mail transport daemons, mailing list software, and spam filters, as well as various other software related to electronic mail.");
-  section_descriptions["math"]=W_("Numeric analysis and other mathematics-related software\n Packages in the 'math' section include calculators, languages for mathematical computation (similar to Mathematica), symbolic algebra packages, and programs to visualize mathematical objects.");
-  section_descriptions["misc"]=W_("Miscellaneous software\n Packages in the 'misc' section have too unusual a function to be classified.");
-  section_descriptions["net"]=W_("Programs to connect to and provide various services\n Packages in the 'net' section include clients and servers for many protocols, tools to manipulate and debug low-level network protocols, IM systems, and other network-related software.");
-  section_descriptions["news"]=W_("Usenet clients and servers\n Packages in the 'news' section are related to the Usenet distributed news system.  They include news readers and news servers.");
-  section_descriptions["oldlibs"]=W_("Obsolete libraries\n Packages in the 'oldlibs' section are obsolete and should not be used by new software.  They are provided for compatibility reasons, or because software distributed by Debian still requires them.\n .\n With very few exceptions, you should not need to explicitly install a package from this section; the package system will install them as required to fulfill dependencies.");
-  section_descriptions["otherosfs"]=W_("Emulators and software to read foreign filesystems\n Packages in the 'otherosfs' section emulate hardware and operating systems and provide tools for transferring data between different operating systems and hardware platforms. (for instance, utilities to read DOS floppies, and utilities to communicate with Palm Pilots)\n .\n It is worth noting that CD burning software is included in THIS section.");
-  section_descriptions["science"]=W_("Software for scientific work\n Packages in the 'science' section include tools for astronomy, biology, and chemistry, as well as other science-related software.");
-  section_descriptions["shells"]=W_("Command shells and alternative console environments\n Packages in the 'shells' section include programs providing a command-line interface.");
-  section_descriptions["sound"]=W_("Utilities to play and record sound\n Packages in the 'sound' section include sound players, recorders, and encoders for many formats, mixers and volume controls, MIDI sequencers and programs to generate musical notation, drivers for sound hardware, and sound processing software.");
-  section_descriptions["tex"]=W_("The TeX typesetting system\n Packages in the 'tex' section are related to TeX, a system for producing high-quality typeset output.  They include TeX itself, TeX packages, editors designed for TeX, utilities to convert TeX and TeX output files to various formats, TeX fonts, and other software related to TeX.");
-  section_descriptions["text"]=W_("Text processing utilities\n Packages in the 'text' section include text filters and processors, spelling checkers, dictionary programs, utilities to convert between character encodings and text file formats (eg, Unix and DOS), text formatters and pretty-printers, and other software which operates on plain text.");
-  section_descriptions["utils"]=W_("Various system utilities\n Packages in the 'utils' section are utilities whose purpose is too unique to be classified.");
-  section_descriptions["web"]=W_("Web browsers, servers, proxies, and other tools\n Packages in the 'web' section include Web browsers, Web servers and proxies, software to write CGI scripts or Web-based programs, pre-written Web-based programs, and other software related to the World Wide Web.");
-  section_descriptions["x11"]=W_("The X window system and related software\n Packages in the 'x11' section include the core packages for the X window system, window managers, utility programs for X, and miscellaneous programs with an X GUI which were placed here because they didn't fit anywhere else.");
-
-  section_descriptions["contrib"]=W_("Programs which depend on software not in Debian\n Packages in the 'contrib' section are not part of Debian.\n .\n These packages are Free Software, but they depend on software which is not part of Debian.  This may be because it is not Free Software, but is packaged in the non-free section of the archive, because Debian cannot distribute it at all, or (in rare cases) because no-one has packaged it yet.\n .\n For more information about what Debian considers to be Free Software, see http://www.debian.org/social_contract#guidelines");
-  section_descriptions["main"]=W_("The main Debian archive\n The Debian distribution consists of packages from the 'main' section. Every package in 'main' is Free Software.\n .\n For more information about what Debian considers to be Free Software, see http://www.debian.org/social_contract#guidelines");
-  section_descriptions["non-US"]=W_("Programs stored outside the US due to export controls\n Packages in 'non-US' likely contain cryptography; a few implement patented algorithms. Because of this, they cannot be exported from the United States, and hence are stored on a server in the \"free world\".\n .\n Note: the Debian Project is currently merging cryptographic software into the US-based archives after consulting with legal experts about recent changes in export policies. Most packages which were formerly found in this section, therefore, are now in 'main'.");
-  section_descriptions["non-free"]=W_("Programs which are not free software\n Packages in the 'non-free' section are not part of Debian.\n .\n These packages fail to meet one or more of the requirements of the Debian Free Software Guidelines (see below). You should read the license of programs in this section to be sure that you are allowed to use them in the way you intend.\n .\n For more information about what Debian considers to be Free Software, see http://www.debian.org/social_contract#guidelines");
-
-  section_descriptions["virtual"]=W_("Virtual packages\n These packages do not exist; they are names other packages use to require or provide some functionality.");
-
-  already_done=true;
+  already_done = true;
 }
 
 void pkg_grouppolicy_section::add_package(const pkgCache::PkgIterator &pkg,
 					  pkg_subtree *root)
 {
-  const char *section=pkg.VersionList().Section();
-  const char *name=pkg.Name();
-  bool maypassthrough=false; // FIXME: HACK!
+  // This flag tracks whether we're in a branch of the logic in which
+  // the passthrough option is obeyed.  That basically means that the
+  // package is virtual, has no section, or is a task package.
+  bool may_passthrough = false;
 
-  if(name[0] == 't' && name[1] == 'a' && name[2] == 's' && name[3] == 'k' && name[4] == '-')
+  string section;
+  if(!strncmp(pkg.Name(), "task-", 5))
     {
-      maypassthrough=true;
-      section=split_mode!=pkg_grouppolicy_section_factory::split_none?_("Tasks/Tasks"):_("Tasks");
+      section=_("Tasks");
+      may_passthrough = true;
     }
-
-  if(!section)
+  else if(pkg.VersionList().end())
     {
-      maypassthrough=true;
-      section=split_mode!=pkg_grouppolicy_section_factory::split_none?_("Unknown/Unknown"):_("Unknown");
+      section=_("virtual");
+      may_passthrough = true;
     }
-
-  if(pkg.VersionList().end())
+  else if(!pkg.VersionList().Section() ||
+	  (*pkg.VersionList().Section()) == '\0')
     {
-      maypassthrough=true;
-      section=split_mode!=pkg_grouppolicy_section_factory::split_none?_("virtual/virtual"):_("virtual");
-    }
-
-  const char *split=strchr(section, '/');
-  const char *subdir=split?split+1:section;
-
-  string tag;
-
-  if(split_mode==pkg_grouppolicy_section_factory::split_none)
-    tag=section;
-  else if(split_mode==pkg_grouppolicy_section_factory::split_topdir)
-    tag=(split?string(section,split-section):string(_("main")));
-  else if(split_mode==pkg_grouppolicy_section_factory::split_subdir)
-    tag=subdir;
-
-  section_map::iterator found=sections.find(tag);
-
-  if(maypassthrough && passthrough)
-    {
-      if(found==sections.end())
-	sections[tag].first=chain->instantiate(get_sig(), get_desc_sig());
-
-      sections[tag].first->add_package(pkg, root);
+      section=_("Unknown");
+      may_passthrough = true;
     }
   else
     {
-      if(found==sections.end())
+      section=pkg.VersionList().Section();
+
+      // Find the first section divider ('/'); if the split mode is
+      // supposed to include only the part of the section preceding it
+      // or only the part following it, modify the section
+      // accordingly.
+      string::size_type first_split = section.find('/');
+      if(split_mode == pkg_grouppolicy_section_factory::split_topdir)
+	section = (first_split != section.npos
+		   ? section.substr(0,first_split)
+		   : _("main"));
+      else if((split_mode == pkg_grouppolicy_section_factory::split_subdir ||
+	       split_mode == pkg_grouppolicy_section_factory::split_subdirs) &&
+	      first_split != section.npos)
+	section = section.substr(first_split + 1);
+    }
+
+  // If passthrough is enabled and this package is in a section that
+  // can pass through, place it directly in the top-level tree.
+  if(passthrough && may_passthrough)
+    {
+      if(sections.find(section) == sections.end())
+	sections[section].first = chain->instantiate(get_sig(), get_desc_sig());
+      sections[section].first->add_package(pkg, root);
+    }
+  // Find the subtree in which this package is to be placed and add it
+  // to that tree.
+  else
+    {
+      pkg_subtree *current_root = root;
+      // The fully qualified path of the tree in which the current
+      // section is to be placed.  "" for the root tree, "games" for
+      // "games" in the root tree, "non-free/games/arcade" for
+      // "arcade" in the tree for "games" in the tree for "non-free"
+      // in the root tree, etc.
+      string section_id;
+      // The portion of the section that hasn't been instantiated in
+      // the tree yet.
+      string sections_remaining = section;
+      // Sentinel value, set to "true" when we've processed the whole
+      // string.
+      bool done = false;
+      do
 	{
-	  pkg_subtree *newtree;
-	  string realtag=tag;
-
-	  // Go by the last element of the section for multi-level sections.
-	  if(tag.find('/')!=tag.npos)
-	    realtag=string(tag, tag.rfind('/')+1);
-
-	  if(section_descriptions.find(realtag)!=section_descriptions.end())
+	  if(split_mode == pkg_grouppolicy_section_factory::split_subdirs)
 	    {
-	      wstring desc=section_descriptions[realtag];
-
-	      if(desc.find(L'\n')!=desc.npos)
-		newtree=new pkg_subtree(cw::util::transcode(tag)+L" - "+wstring(desc, 0, desc.find('\n')),
-					desc,
-					get_desc_sig());
+	      // If we're splitting into subdirectories, we need to
+	      // strip one path element from the string and
+	      // instantiate it.
+	      string::size_type next_split = sections_remaining.find('/', 1);
+	      section_id.append(sections_remaining.substr(0, next_split));
+	      section = (sections_remaining.at(0) == '/'
+			 ? sections_remaining.substr(1, next_split)
+			 : sections_remaining.substr(0, next_split));
+	      if(next_split == sections_remaining.npos)
+		done = true;
 	      else
-		newtree=new pkg_subtree(cw::util::transcode(tag)+desc);
+		sections_remaining = sections_remaining.substr(next_split);
 	    }
 	  else
-	    newtree=new pkg_subtree(cw::util::transcode(tag));
+	    {
+	      // In this case we've precomputed the full path above,
+	      // and the section will be placed in the root node.
+	      section_id = section;
+	      done = true;
+	    }
 
-	  sections[tag].first=chain->instantiate(get_sig(), get_desc_sig());
-	  sections[tag].second=newtree;
+	  // If the subtree into which this package should be placed
+	  // doesn't exist, create it.
+	  section_map::iterator found = sections.find(section_id);
+	  if(found == sections.end())
+	    {
+	      string section_tail = section;
+	      pkg_subtree *newtree;
 
-	  root->add_child(newtree);
-	}
+	      // Look up the description of the section based on the
+	      // last component of its name, and create a new subtree
+	      // for it.
+	      string::size_type last_split = section.rfind('/');
+	      if(last_split != section.npos)
+		section_tail = section.substr(last_split+1);
 
-      sections[tag].first->add_package(pkg, sections[tag].second);
+	      if(section_descriptions.find(section_tail) != section_descriptions.end())
+		{
+		  wstring desc = section_descriptions[section_tail];
+		  if(desc.find(L'\n') != desc.npos)
+		    newtree = new pkg_subtree(cw::util::transcode(section) + L" - " + wstring(desc, 0, desc.find('\n')), desc, get_desc_sig());
+		  else
+		    newtree = new pkg_subtree(cw::util::transcode(section) + desc,
+					      L"", get_desc_sig());
+		}
+	      else
+		newtree = new pkg_subtree(cw::util::transcode(section),
+					  L"", get_desc_sig());
+
+	      // Generate a new sub-grouping-policy, and insert it
+	      // into the map with the new tree.
+	      //
+	      // Note that the new grouping policy is only used for
+	      // packages that are in this section; packages in
+	      // sub-sections go directly to the policy of the
+	      // sub-section.
+	      sections[section_id].first = chain->instantiate(get_sig(), get_desc_sig());
+	      sections[section_id].second = newtree;
+
+	      current_root->add_child(newtree);
+	      newtree->set_num_packages_parent(current_root);
+	    }
+	  current_root=sections[section_id].second;
+	} while(!done);
+
+      sections[section_id].first->add_package(pkg, sections[section_id].second);
     }
 }
 
@@ -396,6 +490,7 @@ void pkg_grouppolicy_status::add_package(const pkgCache::PkgIterator &pkg,
       children[section].first=chain->instantiate(get_sig(), get_desc_sig());
       children[section].second=newtree;
       root->add_child(newtree);
+      newtree->set_num_packages_parent(root);
     }
 
   children[section].first->add_package(pkg, children[section].second);
@@ -404,11 +499,11 @@ void pkg_grouppolicy_status::add_package(const pkgCache::PkgIterator &pkg,
 /*****************************************************************************/
 class pkg_grouppolicy_filter:public pkg_grouppolicy
 {
-  pkg_matcher *filter;
+  match::pkg_matcher *filter;
 
   pkg_grouppolicy *chain;
 public:
-  pkg_grouppolicy_filter(pkg_grouppolicy_factory *_chain, pkg_matcher *_filter,
+  pkg_grouppolicy_filter(pkg_grouppolicy_factory *_chain, match::pkg_matcher *_filter,
 			 pkg_signal *_sig, desc_signal *_desc_sig)
     :pkg_grouppolicy(_sig, _desc_sig),
      filter(_filter),
@@ -418,7 +513,7 @@ public:
 
   virtual void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root)
   {
-    if(filter->matches(pkg, *apt_cache_file, *apt_package_records))
+    if(match::apply_matcher(filter, pkg, *apt_cache_file, *apt_package_records))
       chain->add_package(pkg, root);
   }
 
@@ -474,6 +569,7 @@ public:
 							    group,
 							    true);
 	    root->add_child(newtree);
+	    newtree->set_num_packages_parent(root);
 	    children[group].first=chain->instantiate(get_sig(),
 						     get_desc_sig());
 	    children[group].second=newtree;
@@ -494,6 +590,7 @@ public:
 							    num_pkg_action_states,
 							    true);
 	    root->add_child(newtree);
+	    newtree->set_num_packages_parent(root);
 	    recommended_child.first=chain->instantiate(get_sig(),
 						       get_desc_sig());
 	    recommended_child.second=newtree;
@@ -514,6 +611,7 @@ public:
 							    num_pkg_action_states+1,
 							    false);
 	    root->add_child(newtree);
+	    newtree->set_num_packages_parent(root);
 	    suggested_child.first=chain->instantiate(get_sig(),
 						     get_desc_sig());
 	    suggested_child.second=newtree;
@@ -588,12 +686,14 @@ public:
 	string treename;
 	treename+=firstchar;
 
-	pkg_subtree *newtree=new pkg_subtree(cw::util::transcode(treename));
+	pkg_subtree *newtree=new pkg_subtree(cw::util::transcode(treename),
+					     L"", get_desc_sig());
 	pkg_grouppolicy *newchild=chain->instantiate(get_sig(),
 						     get_desc_sig());
 	children[firstchar].first=newchild;
 	children[firstchar].second=newtree;
 	root->add_child(newtree);
+	newtree->set_num_packages_parent(root);
 
 	newchild->add_package(pkg, newtree);
       }
@@ -681,12 +781,16 @@ public:
 	  default                        : order=6; break;
 	  };
 
-	pkg_subtree *newtree=new pkg_subtree_with_order(cw::util::transcode(buf), order);
+	pkg_subtree *newtree=new pkg_subtree_with_order(cw::util::transcode(buf),
+							L"",
+							get_desc_sig(),
+							order);
 	pkg_grouppolicy *newchild=chain->instantiate(get_sig(),
 						     get_desc_sig());
 	children[priority].first=newchild;
 	children[priority].second=newtree;
 	root->add_child(newtree);
+	newtree->set_num_packages_parent(root);
 
 	newchild->add_package(pkg, newtree);
       }
@@ -779,9 +883,10 @@ public:
       {
 	if(!uncategorized)
 	  {
-	    uncategorized=new pkg_subtree(W_("UNCATEGORIZED"));
+	    uncategorized=new pkg_subtree(W_("UNCATEGORIZED"), L"", desc_sig);
 	    uncategorized_policy=chain->instantiate(sig, desc_sig);
 	    root->add_child(uncategorized);
+	    uncategorized->set_num_packages_parent(root);
 	  }
 
 	uncategorized_policy->add_package(pkg, uncategorized);
@@ -792,16 +897,23 @@ public:
   {
     nodedata *data=(nodedata *) parent_data;
 
-    pkg_subtree *newtree=new pkg_subtree(cw::util::transcode(group->description));
+    pkg_subtree *newtree=new pkg_subtree(cw::util::transcode(group->description),
+					 L"", desc_sig);
     pkg_grouppolicy *newpolicy=chain->instantiate(sig, desc_sig);
 
     nodedata *rval=new nodedata(newpolicy, newtree);
     data_out_there.push_back(rval);
 
     if(data)
-      data->second->add_child(newtree);
+      {
+	data->second->add_child(newtree);
+	newtree->set_num_packages_parent(data->second);
+      }
     else
-      root->add_child(newtree);
+      {
+	root->add_child(newtree);
+	newtree->set_num_packages_parent(root);
+      }
 
     return rval;
   }
@@ -857,7 +969,7 @@ pkg_grouppolicy_hier_factory::~pkg_grouppolicy_hier_factory()
 class pkg_grouppolicy_task:public pkg_grouppolicy
 {
   /** The subtrees are indexed by Task. */
-  typedef hash_map<string, pkg_subtree *> subtree_map;
+  typedef std::map<string, pkg_subtree *> subtree_map;
 
   /** A list of [task,description] pairs. */
   typedef vector<pair<string, wstring> > desclist;
@@ -871,7 +983,7 @@ class pkg_grouppolicy_task:public pkg_grouppolicy
 
   static desclist task_section_descriptions;
   // Lifted from tasksel.  There are so few of them that it's not
-  // worth using a hash_map.  Even a map is overkill.
+  // worth using a std::map, IMNSHO.
 
   static void init_section_descriptions();
 public:
@@ -927,13 +1039,13 @@ public:
 void pkg_grouppolicy_task::add_package(const pkgCache::PkgIterator &pkg,
 				       pkg_subtree *root)
 {
-  list<string> *tasks=get_tasks(pkg);
+  set<string> *tasks=get_tasks(pkg);
 
   eassert(tasks);
 
   chain->add_package(pkg, root);
 
-  for(list<string>::iterator i=tasks->begin(); i!=tasks->end(); ++i)
+  for(set<string>::iterator i = tasks->begin(); i != tasks->end(); ++i)
     {
       subtree_map::iterator found=task_children.find(*i);
 
@@ -958,6 +1070,7 @@ void pkg_grouppolicy_task::add_package(const pkgCache::PkgIterator &pkg,
 	      tasks_subtree=new pkg_subtree(W_("Tasks"),
 					    W_("\n Tasks are groups of packages which provide an easy way to select a predefined set of packages for a particular purpose."), get_desc_sig());
 	      root->add_child(tasks_subtree);
+	      tasks_subtree->set_num_packages_parent(root);
 	    }
 
 	  subtree_map::iterator sectionfound=section_children.find(section);
@@ -974,9 +1087,10 @@ void pkg_grouppolicy_task::add_package(const pkgCache::PkgIterator &pkg,
 		    break;
 		  }
 
-	      sectiontree=new pkg_subtree(sectiondesc);
+	      sectiontree = new pkg_subtree(sectiondesc, L"", get_desc_sig());
 	      section_children[section]=sectiontree;
 	      tasks_subtree->add_child(sectiontree);
+	      sectiontree->set_num_packages_parent(tasks_subtree);
 	    }
 	  else
 	    sectiontree=sectionfound->second;
@@ -993,11 +1107,16 @@ void pkg_grouppolicy_task::add_package(const pkgCache::PkgIterator &pkg,
 	  task_children[*i]=newtree;
 
 	  sectiontree->add_child(newtree);
+	  newtree->set_num_packages_parent(sectiontree);
 
 	  newtree->add_child(new pkg_item(pkg, get_sig()));
+	  newtree->inc_num_packages();
 	}
       else
-	found->second->add_child(new pkg_item(pkg, get_sig()));
+	{
+	  found->second->add_child(new pkg_item(pkg, get_sig()));
+	  found->second->inc_num_packages();
+	}
     }
 }
 
@@ -1040,7 +1159,7 @@ private:
   subtree_map subtrees;
 
   wstring substitute(const wstring &s,
-		     pkg_match_result *res)
+		     match::pkg_match_result *res)
   {
     wstring rval;
 
@@ -1144,7 +1263,7 @@ public:
     for(vector<match_pair>::const_iterator i = subgroups.begin();
 	i != subgroups.end(); ++i)
 	{
-	  pkg_match_result *res = i->matcher->get_match(pkg, *apt_cache_file, *apt_package_records);
+	  match::pkg_match_result *res = match::get_match(i->matcher, pkg, *apt_cache_file, *apt_package_records);
 	  if(res != NULL)
 	    {
 	      if(i->passthrough)
@@ -1166,10 +1285,11 @@ public:
 		found->second.policy->add_package(pkg, found->second.tree);
 	      else
 		{
-		  pkg_subtree *tree = new pkg_subtree(title);
+		  pkg_subtree *tree = new pkg_subtree(title, L"", get_desc_sig());
 		  pkg_grouppolicy *policy = chain->instantiate(get_sig(),
 							       get_desc_sig());
 		  root->add_child(tree);
+		  tree->set_num_packages_parent(root);
 
 		  subtrees[title]=subtree_pair(policy, tree);
 
@@ -1206,7 +1326,7 @@ class pkg_grouppolicy_tag : public pkg_grouppolicy
 
   string facet;
 
-  typedef hash_map<string, pair<pkg_grouppolicy *, pkg_subtree *> > childmap;
+  typedef std::map<string, pair<pkg_grouppolicy *, pkg_subtree *> > childmap;
 
   childmap children;
 public:
@@ -1219,12 +1339,39 @@ public:
   virtual void add_package(const pkgCache::PkgIterator &pkg,
 			   pkg_subtree *root)
   {
-    const set<tag> *tags = get_tags(pkg);
-    if(tags == NULL)
+#ifdef HAVE_EPT
+    typedef ept::debtags::Tag tag;
+    using aptitude::apt::get_tags;
+#endif
+
+#ifdef HAVE_EPT
+    const set<tag> realTags(get_tags(pkg));
+    const set<tag> * const tags(&realTags);
+#else
+    const set<tag> * const tags(get_tags(pkg));
+#endif
+
+    if(tags != NULL)
       return;
+
     for(set<tag>::const_iterator ti = tags->begin();
 	ti != tags->end(); ++ti)
       {
+#ifdef HAVE_EPT
+	const ept::debtags::Facet f = ti->facet();
+
+	const std::string thisfacet(f.name());
+
+	// Don't create items for tags that aren't in our facet.
+	if(f.name() != facet)
+	  return;
+
+	// TODO: split up by sub-facet or whatever the debtags guys
+	// call it?
+	std::string tagname;
+	ti->name(tagname);
+
+#else // HAVE_EPT
 	tag::const_iterator j = ti->begin();
 	if(j == ti->end())
 	  continue;
@@ -1239,6 +1386,8 @@ public:
 	  continue;
 
 	string tagname = *j;
+#endif // HAVE_EPT
+
 	childmap::const_iterator found =
 	  children.find(tagname);
 
@@ -1247,8 +1396,16 @@ public:
 
 	if(found == children.end())
 	  {
+#ifdef HAVE_EPT
+	    std::string desc;
+	    ti->longDescription(desc);
+
+	    std::string shortdesc;
+	    ti->shortDescription(shortdesc);
+#else // HAVE_EPT
 	    string desc = tag_description(ti->str());
 	    string shortdesc(desc, 0, desc.find('\n'));
+#endif
 
 	    subtree = new pkg_subtree(swsprintf(L"%s - %s",
 						tagname.c_str(),
@@ -1256,6 +1413,7 @@ public:
 				      cw::util::transcode(desc),
 				      get_desc_sig());
 	    root->add_child(subtree);
+	    subtree->set_num_packages_parent(root);
 	    subpolicy = chain->instantiate(get_sig(),
 					   get_desc_sig());
 
@@ -1299,9 +1457,9 @@ class pkg_grouppolicy_facet_tag : public pkg_grouppolicy
 {
   pkg_grouppolicy_factory *chain;
 
-  typedef hash_map<string, pair<pkg_grouppolicy *, pkg_subtree *> > tagmap;
+  typedef std::map<string, pair<pkg_grouppolicy *, pkg_subtree *> > tagmap;
 
-  typedef hash_map<string, pair<tagmap *, pkg_subtree *> > facetmap;
+  typedef std::map<string, pair<tagmap *, pkg_subtree *> > facetmap;
 
   pkg_subtree *untagged_tree;
   pkg_grouppolicy *untagged_policy;
@@ -1322,22 +1480,30 @@ public:
   virtual void add_package(const pkgCache::PkgIterator &pkg,
 			   pkg_subtree *root)
   {
-    const set<tag> *tags = get_tags(pkg);
+#ifdef HAVE_EPT
+    typedef ept::debtags::Tag tag;
+    using aptitude::apt::get_tags;
+#endif
 
-    if(tags == NULL)
-      return;
+#ifdef HAVE_EPT
+    const set<tag> realTags(get_tags(pkg));
+    const set<tag> * const tags(&realTags);
+#else
+    const set<tag> * const tags(get_tags(pkg));
+#endif
 
     // Put all untagged, non-virtual packages into a separate list.
-    if(tags->empty() && !pkg.VersionList().end())
+    if((tags == NULL || tags->empty()) && !pkg.VersionList().end())
       {
 	if(untagged_tree == NULL)
 	  {
 	    eassert(untagged_policy == NULL);
 
 	    untagged_tree = new pkg_subtree(W_("TAGLESS PACKAGES"),
-					    W_("\n These packages have not yet been classified in debtags."),
+					    W_("\n These packages have not yet been classified in debtags, or the debtags database is not present (installing debtags may correct this problem)."),
 					    get_desc_sig());
 	    root->add_child(untagged_tree);
+	    untagged_tree->set_num_packages_parent(root);
 
 	    untagged_policy = chain->instantiate(get_sig(), get_desc_sig());
 	  }
@@ -1345,9 +1511,17 @@ public:
 	untagged_policy->add_package(pkg, untagged_tree);
       }
 
+    if(tags == NULL)
+      return;
+
     for(set<tag>::const_iterator ti = tags->begin();
 	ti != tags->end(); ++ti)
       {
+#ifdef HAVE_EPT
+	const ept::debtags::Facet f(ti->facet());
+	std::string thisfacet(f.name());
+	std::string thistag(ti->name());
+#else // HAVE_EPT
 	tag::const_iterator j = ti->begin();
 
 	eassert(j != ti->end());
@@ -1363,6 +1537,7 @@ public:
 	  thistag = _("MISSING TAG");
 	else
 	  thistag = *j;
+#endif
 
 	facetmap::const_iterator facetfound =
 	  children.find(thisfacet);
@@ -1372,8 +1547,13 @@ public:
 
 	if(facetfound == children.end())
 	  {
+#ifdef HAVE_EPT
+	    string desc(f.longDescription());
+	    string shortdesc(f.shortDescription());
+#else // HAVE_EPT
 	    string desc = facet_description(thisfacet);
 	    string shortdesc(desc, 0, desc.find('\n'));
+#endif
 
 	    if(!shortdesc.empty())
 	      tagtree = new pkg_subtree(swsprintf(L"%s - %s",
@@ -1386,6 +1566,7 @@ public:
 					cw::util::transcode(desc),
 					get_desc_sig());
 	    root->add_child(tagtree);
+	    tagtree->set_num_packages_parent(root);
 	    tagchildren = new tagmap;
 
 	    children[thisfacet] = pair<tagmap *, pkg_subtree *>(tagchildren, tagtree);
@@ -1404,8 +1585,13 @@ public:
 
 	if(tagfound == tagchildren->end())
 	  {
+#ifdef HAVE_EPT
+	    string desc(ti->longDescription());
+	    string shortdesc(ti->shortDescription());
+#else // HAVE_EPT
 	    string desc = tag_description(ti->str());
 	    string shortdesc(desc, 0, desc.find('\n'));
+#endif
 
 	    if(!shortdesc.empty())
 	      subtree = new pkg_subtree(swsprintf(L"%s - %s",
@@ -1419,6 +1605,7 @@ public:
 					get_desc_sig());
 
 	    tagtree->add_child(subtree);
+	    subtree->set_num_packages_parent(tagtree);
 	    subpolicy = chain->instantiate(get_sig(), get_desc_sig());
 
 	    (*tagchildren)[thistag] = pair<pkg_grouppolicy *, pkg_subtree *>(subpolicy, subtree);
