@@ -1,6 +1,6 @@
 // ui.cc
 //
-//   Copyright 2000-2007 Daniel Burrows <dburrows@debian.org>
+//   Copyright 2000-2008 Daniel Burrows <dburrows@debian.org>
 //
 //   This program is free software; you can redistribute it and/or
 //   modify it under the terms of the GNU General Public License as
@@ -32,7 +32,6 @@
 #include <apt-pkg/error.h>
 #include <apt-pkg/packagemanager.h>
 
-#include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <sys/wait.h>
@@ -100,9 +99,7 @@
 #include <generic/util/util.h>
 
 #include "dep_item.h"
-#include "download_bar.h"
 #include "download_list.h"
-#include "download_screen.h"
 #include "menu_tree.h"
 #include "pkg_columnizer.h"
 #include "pkg_grouppolicy.h"
@@ -158,6 +155,7 @@ sigc::signal0<void> package_states_changed;
 sigc::signal0<bool, cw::util::accumulate_or> package_menu_enabled;
 sigc::signal0<bool, cw::util::accumulate_or> package_forbid_enabled;
 sigc::signal0<bool, cw::util::accumulate_or> package_information_enabled;
+sigc::signal0<bool, cw::util::accumulate_or> package_cycle_information_enabled;
 sigc::signal0<bool, cw::util::accumulate_or> package_changelog_enabled;
 
 sigc::signal0<bool, cw::util::accumulate_or> find_search_enabled;
@@ -177,6 +175,7 @@ sigc::signal0<bool, cw::util::accumulate_or> package_mark_auto;
 sigc::signal0<bool, cw::util::accumulate_or> package_unmark_auto;
 sigc::signal0<bool, cw::util::accumulate_or> package_forbid;
 sigc::signal0<bool, cw::util::accumulate_or> package_information;
+sigc::signal0<bool, cw::util::accumulate_or> package_cycle_information;
 sigc::signal0<bool, cw::util::accumulate_or> package_changelog;
 
 sigc::signal0<bool, cw::util::accumulate_or> resolver_toggle_rejected;
@@ -194,9 +193,12 @@ sigc::signal0<bool, cw::util::accumulate_or> find_limit;
 sigc::signal0<bool, cw::util::accumulate_or> find_cancel_limit;
 sigc::signal0<bool, cw::util::accumulate_or> find_broken;
 
+sigc::signal1<void, bool> install_finished;
+sigc::signal1<void, bool> update_finished;
+
 const char *default_pkgstatusdisplay="%d";
 const char *default_pkgheaderdisplay="%N %n #%B %u %o";
-const char *default_grpstr="task,status,section(subdir,passthrough),section(topdir)";
+const char *default_grpstr="task,status,section(subdirs,passthrough),section(topdir)";
 
 void ui_start_download(bool hide_preview)
 {
@@ -821,7 +823,7 @@ void do_new_package_view(OpProgress &progress)
       if(!grp)
 	// Eek!  The default grouping failed to parse.  Fall all the
 	// way back.
-	grp=new pkg_grouppolicy_task_factory(new pkg_grouppolicy_status_factory(new pkg_grouppolicy_section_factory(pkg_grouppolicy_section_factory::split_subdir,true,new pkg_grouppolicy_section_factory(pkg_grouppolicy_section_factory::split_topdir,false,new pkg_grouppolicy_end_factory()))));
+	grp=new pkg_grouppolicy_task_factory(new pkg_grouppolicy_status_factory(new pkg_grouppolicy_section_factory(pkg_grouppolicy_section_factory::split_subdirs,true,new pkg_grouppolicy_section_factory(pkg_grouppolicy_section_factory::split_topdir,false,new pkg_grouppolicy_end_factory()))));
     }
 
   pkg_tree_ref tree=pkg_tree::create(grpstr.c_str(), grp);
@@ -849,7 +851,7 @@ static void do_new_recommendations_view()
   progress_ref p = gen_progress_bar();
 
   pkg_grouppolicy_factory *grp = new pkg_grouppolicy_end_factory();
-  std::string grpstr="section(subdir, passthrough)";
+  std::string grpstr="section(subdirs, passthrough)";
 
   pkg_tree_ref tree=pkg_tree::create(grpstr.c_str(), grp,
 				     L"!~v!~i~RBrecommends:~i");
@@ -965,7 +967,9 @@ cw::widget_ref make_ver_screen(const pkgCache::PkgIterator &pkg)
 
 static void do_help_about()
 {
-  cw::fragment *f=cw::fragf(_("Aptitude %s%n%nCopyright 2000-2005 Daniel Burrows.%n%naptitude comes with %BABSOLUTELY NO WARRANTY%b; for details see 'license' in the Help menu.  This is free software, and you are welcome to redistribute it under certain conditions; see 'license' for details."), VERSION);
+  cw::fragment *f = cw::fragf(_("Aptitude %s%n%nCopyright 2000-2008 Daniel Burrows.%n"
+				"%n"
+				"aptitude comes with %BABSOLUTELY NO WARRANTY%b; for details see 'license' in the Help menu.  This is free software, and you are welcome to redistribute it under certain conditions; see 'license' for details."), VERSION);
 
   cw::widget_ref w=cw::dialogs::ok(wrapbox(f));
   w->show_all();
@@ -1083,65 +1087,13 @@ static void do_help_news()
   setup_fileview(HELPDIR "/NEWS",
 		 NULL,
 		 _("News"),
-		 _("View the important changes made in each version of " PACKAGE),
+		 ssprintf(_("View the important changes made in each version of %s"), PACKAGE),
 		 _("News"));
 }
 
-/** Remove the given file/directory and all its children.  Behaves
- *  like rm -fr.
- *
- *  \todo this belongs in util.cc
- */
-static bool recursive_remdir(const std::string &dirname)
-{
-  struct stat stbuf;
-
-  if(lstat(dirname.c_str(), &stbuf) != 0)
-    _error->Errno("recursive_rmdir", _("Unable to stat \"%s\""), dirname.c_str());
-
-  if(S_ISLNK(stbuf.st_mode) || !S_ISDIR(stbuf.st_mode))
-    {
-      if(unlink(dirname.c_str()) != 0)
-	{
-	  _error->Errno("recursive_rmdir", _("Unable to remove \"%s\""), dirname.c_str());
-	  return false;
-	}
-      else
-	return true;
-    }
-
-  DIR *dir = opendir(dirname.c_str());
-  if(dir == NULL)
-    {
-      _error->Errno("recursive_rmdir", _("Unable to list files in \"%s\""), dirname.c_str());
-      return false;
-    }
-
-  bool rval = true;
-
-  for(dirent *dent = readdir(dir); dent != NULL; dent = readdir(dir))
-    if(strcmp(dent->d_name, ".") != 0 &&
-       strcmp(dent->d_name, "..") != 0)
-      rval = (rval && recursive_remdir(dirname + "/" + dent->d_name));
-
-  if(closedir(dir) != 0)
-    {
-      _error->Errno("recursive_rmdir", _("Failure closing directory \"%s\""), dirname.c_str());
-      rval = false;
-    }
-
-  if(rmdir(dirname.c_str()) != 0)
-    {
-      _error->Errno("recursive_rmdir", _("Unable to remove directory \"%s\""), dirname.c_str());
-      rval = false;
-    }
-
-  return rval;
-}
-
 static void do_kill_old_tmp(const std::string old_tmpdir)
 {
-  if(!recursive_remdir(old_tmpdir))
+  if(!aptitude::util::recursive_remdir(old_tmpdir))
     show_message(ssprintf(_("Unable to remove the old temporary directory; you should remove %s by hand."), old_tmpdir.c_str()));
 }
 
@@ -1214,10 +1166,14 @@ void install_or_remove_packages()
   m->post_install_hook.connect(sigc::ptr_fun(&finish_install_run));
   m->post_forget_new_hook.connect(package_states_changed.make_slot());
 
-  (new ui_download_manager(m, false, false, true,
-			   _("Downloading packages"),
-			   _("View the progress of the package download"),
-			   _("Package Download")))->start();
+  ui_download_manager *uim =
+    new ui_download_manager(m, false, false, true,
+			    _("Downloading packages"),
+			    _("View the progress of the package download"),
+			    _("Package Download"));
+
+  uim->download_complete.connect(install_finished.make_slot());
+  uim->start();
 }
 
 /** Make sure that no trust violations are about to be committed.  If
@@ -1559,10 +1515,14 @@ void really_do_update_lists()
   m->pre_autoclean_hook.connect(sigc::bind(sigc::ptr_fun(lists_autoclean_msg),
 					   m));
   m->post_forget_new_hook.connect(package_states_changed.make_slot());
-  (new ui_download_manager(m, false, true, false,
-			   _("Updating package lists"),
-			   _("View the progress of the package list update"),
-			   _("List Update")))->start();
+  ui_download_manager *uim =
+    new ui_download_manager(m, false, true, false,
+			    _("Updating package lists"),
+			    _("View the progress of the package list update"),
+			    _("List Update"));
+
+  uim->download_complete.connect(update_finished.make_slot());
+  uim->start();
 }
 
 void do_update_lists()
@@ -1854,12 +1814,16 @@ static void start_solution_calculation(bool blocking)
       const int selected = state.selected_solution;
       const int limit = aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 5000);
       const int wait_steps = aptcfg->FindI(PACKAGE "::ProblemResolver::WaitSteps", 50);
-      interactive_continuation * const k = new interactive_continuation(resman);
 
-      if(blocking)
-	resman->get_solution_background_blocking(selected, limit, wait_steps, k);
-      else
-	resman->get_solution_background(selected, limit, k);
+      if(limit > 0)
+	{
+	  interactive_continuation * const k = new interactive_continuation(resman);
+
+	  if(blocking)
+	    resman->get_solution_background_blocking(selected, limit, wait_steps, k);
+	  else
+	    resman->get_solution_background(selected, limit, k);
+	}
     }
 }
 
@@ -2019,6 +1983,22 @@ void do_apply_solution()
     }
 }
 
+namespace
+{
+  bool do_reject_break_holds_enabled()
+  {
+    return resman != NULL && resman->resolver_exists();
+  }
+
+  void do_reject_break_holds()
+  {
+    if(!do_reject_break_holds_enabled())
+      beep();
+    else
+      resman->reject_break_holds();
+  }
+}
+
 static void do_nullify_solver(cw::widget_ref *solver)
 {
   *solver = NULL;
@@ -2026,7 +2006,8 @@ static void do_nullify_solver(cw::widget_ref *solver)
 
 static bool do_examine_solution_enabled()
 {
-  return resman != NULL && resman->resolver_exists();
+  return resman != NULL && resman->resolver_exists() &&
+    aptcfg->FindI(PACKAGE "::ProblemResolver::StepLimit", 5000) > 0;
 }
 
 void do_examine_solution()
@@ -2204,6 +2185,10 @@ cw::menu_info package_menu[]={
 	       N_("Display more information about the selected package"),
 	       sigc::hide_return(package_information.make_slot()),
 	       package_information_enabled.make_slot()),
+  cw::menu_info(cw::menu_info::MENU_ITEM, N_("C^ycle Package Information"), "DescriptionCycle",
+		N_("Cycle through the modes of the package information area: it can show the package's long description, a summary of its dependency status, or an analysis of why the package is required."),
+		sigc::hide_return(package_cycle_information.make_slot()),
+		package_cycle_information_enabled.make_slot()),
   cw::menu_info(cw::menu_info::MENU_ITEM, N_("^Changelog"), "Changelog",
 	       N_("Display the Debian changelog of the selected package"),
 	       sigc::hide_return(package_changelog.make_slot()),
@@ -2254,6 +2239,14 @@ cw::menu_info resolver_menu[] = {
 	       sigc::hide_return(resolver_view_target.make_slot()),
 	       resolver_view_target_enabled.make_slot()),
 
+  cw::menu_info::MENU_SEPARATOR,
+
+  cw::menu_info(cw::menu_info::MENU_ITEM, N_("Reject Breaking ^Holds"),
+		"RejectBreakHolds",
+		N_("Reject all actions that would change the state of held packages or install forbidden versions"),
+		sigc::ptr_fun(&do_reject_break_holds),
+		sigc::ptr_fun(&do_reject_break_holds_enabled)),
+
   cw::menu_info::MENU_END
 };
 
@@ -2381,11 +2374,11 @@ cw::menu_info help_menu_info[]={
 	       sigc::ptr_fun(do_help_faq)),
 
   cw::menu_info(cw::menu_info::MENU_ITEM, N_("^News"), NULL,
-	       N_("View the important changes made in each version of " PACKAGE),
+		N_("View the important changes made in each version of " PACKAGE),
 	       sigc::ptr_fun(do_help_news)),
 
   cw::menu_info(cw::menu_info::MENU_ITEM, N_("^License"), NULL,
-	       N_("View the terms under which you may copy and distribute aptiutde"),
+	       N_("View the terms under which you may copy and distribute aptitude"),
 	       sigc::ptr_fun(do_help_license)),
 
   cw::menu_info::MENU_END
