@@ -2,9 +2,6 @@
 //
 //  Copyright 2000,2001 Daniel Burrows
 //
-//  Global UI definitions and routines.  A lot of things here just provide a
-// single point for common actions so those actions can be customized later.
-// (eg, the progress-bar can appear in various forms)
 
 #ifndef UI_H
 #define UI_H
@@ -20,6 +17,16 @@
 
 #include <cwidget/widgets/editline.h> // Included for history support.
 
+/** \brief Global UI definitions and routines
+ *
+ * 
+ *  Global UI definitions and routines.  A lot of things here just provide a
+ *  single point for common actions so those actions can be customized later.
+ *  (eg, the progress-bar can appear in various forms)
+ * 
+ *  \file ui.h
+ */
+
 class OpProgress;
 
 class download_signal_log;
@@ -106,6 +113,12 @@ extern sigc::signal0<bool, cwidget::util::accumulate_or> package_information_ena
 /** Emitted for Package -> Package Information. */
 extern sigc::signal0<bool, cwidget::util::accumulate_or> package_information;
 
+/** Tests whether Package -> Cycle Information is enabled. */
+extern sigc::signal0<bool, cwidget::util::accumulate_or> package_cycle_information_enabled;
+
+/** Emitted for Package -> Cycle Information. */
+extern sigc::signal0<bool, cwidget::util::accumulate_or> package_cycle_information;
+
 /** Tests whether Package -> Changelog is enabled. */
 extern sigc::signal0<bool, cwidget::util::accumulate_or> package_changelog_enabled;
 
@@ -175,6 +188,16 @@ extern sigc::signal0<bool, cwidget::util::accumulate_or> find_broken;
 
 /*****************************************************************************/
 
+/** \brief Emitted when an install completes.  The parameter is \b
+ *  true if the install succeeded and \b false otherwise.
+ */
+extern sigc::signal1<void, bool> install_finished;
+
+/** \brief Emitted when a list update completes.  The parameter is \b
+ *  true if the update succeeded and \b false otherwise.
+ */
+extern sigc::signal1<void, bool> update_finished;
+
 void ui_init();
 void ui_main();
 
